@@ -44,11 +44,10 @@
 #include "list.h"
 #include "debug.h"
 #include "load_dir.h"
-#ifdef HAVE_CONFIG
-#include "config/datadir.h"
-#endif
 #include "help.h"
 #include "op.h"
+#include "mpris.h"
+#include "job.h"
 
 #include <stdlib.h>
 #include <ctype.h>
@@ -60,7 +59,6 @@ static struct history cmd_history;
 static char *cmd_history_filename;
 static char *history_search_text = NULL;
 static int arg_expand_cmd = -1;
-static int prev_view = -1;
 
 /* view {{{ */
 
@@ -69,26 +67,18 @@ void view_clear(int view)
 	switch (view) {
 	case TREE_VIEW:
 	case SORTED_VIEW:
-		worker_remove_jobs(JOB_TYPE_LIB);
-		editable_lock();
+		worker_remove_jobs_by_type(JOB_TYPE_LIB);
 		editable_clear(&lib_editable);
 
 		/* FIXME: make this optional? */
 		lib_clear_store();
-
-		editable_unlock();
 		break;
 	case PLAYLIST_VIEW:
-		worker_remove_jobs(JOB_TYPE_PL);
-		editable_lock();
-		editable_clear(&pl_editable);
-		editable_unlock();
+		pl_clear();
 		break;
 	case QUEUE_VIEW:
-		worker_remove_jobs(JOB_TYPE_QUEUE);
-		editable_lock();
+		worker_remove_jobs_by_type(JOB_TYPE_QUEUE);
 		editable_clear(&pq_editable);
-		editable_unlock();
 		break;
 	default:
 		info_msg(":clear only works in views 1-4");
@@ -112,16 +102,18 @@ void view_add(int view, char *arg, int prepend)
 	switch (view) {
 	case TREE_VIEW:
 	case SORTED_VIEW:
-		cmus_add(lib_add_track, name, ft, JOB_TYPE_LIB, 0);
+		cmus_add(lib_add_track, name, ft, JOB_TYPE_LIB, 0, NULL);
 		break;
 	case PLAYLIST_VIEW:
-		cmus_add(pl_add_track, name, ft, JOB_TYPE_PL, 0);
+		pl_add_file_to_marked_pl(name);
 		break;
 	case QUEUE_VIEW:
 		if (prepend) {
-			cmus_add(play_queue_prepend, name, ft, JOB_TYPE_QUEUE, 0);
+			cmus_add(play_queue_prepend, name, ft, JOB_TYPE_QUEUE,
+					0, NULL);
 		} else {
-			cmus_add(play_queue_append, name, ft, JOB_TYPE_QUEUE, 0);
+			cmus_add(play_queue_append, name, ft, JOB_TYPE_QUEUE, 0,
+					NULL);
 		}
 		break;
 	default:
@@ -130,17 +122,14 @@ void view_add(int view, char *arg, int prepend)
 	free(name);
 }
 
-void view_load(int view, char *arg)
+static char *view_load_prepare(char *arg)
 {
-	char *tmp, *name;
-	enum file_type ft;
-
-	tmp = expand_filename(arg);
-	ft = cmus_detect_ft(tmp, &name);
+	char *name, *tmp = expand_filename(arg);
+	enum file_type ft = cmus_detect_ft(tmp, &name);
 	if (ft == FILE_TYPE_INVALID) {
 		error_msg("loading '%s': %s", tmp, strerror(errno));
 		free(tmp);
-		return;
+		return NULL;
 	}
 	free(tmp);
 
@@ -149,36 +138,35 @@ void view_load(int view, char *arg)
 	if (ft != FILE_TYPE_PL) {
 		error_msg("loading '%s': not a playlist file", name);
 		free(name);
-		return;
+		return NULL;
 	}
+	return name;
+}
+
+void view_load(int view, char *arg)
+{
+	char *name = view_load_prepare(arg);
+	if (!name)
+		return;
 
 	switch (view) {
 	case TREE_VIEW:
 	case SORTED_VIEW:
-		worker_remove_jobs(JOB_TYPE_LIB);
-		editable_lock();
+		worker_remove_jobs_by_type(JOB_TYPE_LIB);
 		editable_clear(&lib_editable);
-		editable_unlock();
-		cmus_add(lib_add_track, name, FILE_TYPE_PL, JOB_TYPE_LIB, 0);
+		cmus_add(lib_add_track, name, FILE_TYPE_PL, JOB_TYPE_LIB, 0,
+				NULL);
 		free(lib_filename);
 		lib_filename = name;
 		break;
-	case PLAYLIST_VIEW:
-		worker_remove_jobs(JOB_TYPE_PL);
-		editable_lock();
-		editable_clear(&pl_editable);
-		editable_unlock();
-		cmus_add(pl_add_track, name, FILE_TYPE_PL, JOB_TYPE_PL, 0);
-		free(pl_filename);
-		pl_filename = name;
-		break;
 	default:
-		info_msg(":load only works in views 1-3");
+		info_msg(":load only works in views 1-2");
 		free(name);
 	}
 }
 
-static void do_save(for_each_ti_cb for_each_ti, const char *arg, char **filenamep, save_ti_cb save_ti)
+static void do_save(for_each_ti_cb for_each_ti, const char *arg, char **filenamep,
+		save_ti_cb save_ti)
 {
 	char *filename = *filenamep;
 
@@ -195,10 +183,8 @@ static void do_save(for_each_ti_cb for_each_ti, const char *arg, char **filename
 		return;
 	}
 
-	editable_lock();
-	if (save_ti(for_each_ti, filename) == -1)
+	if (save_ti(for_each_ti, filename, NULL) == -1)
 		error_msg("saving '%s': %s", filename, strerror(errno));
-	editable_unlock();
 }
 
 void view_save(int view, char *arg, int to_stdout, int filtered, int extended)
@@ -220,25 +206,22 @@ void view_save(int view, char *arg, int to_stdout, int filtered, int extended)
 	switch (view) {
 	case TREE_VIEW:
 	case SORTED_VIEW:
-		if (worker_has_job(JOB_TYPE_LIB))
+		if (worker_has_job_by_type(JOB_TYPE_LIB))
 			goto worker_running;
 		dest = extended ? &lib_ext_filename : &lib_filename;
 		do_save(lib_for_each_ti, arg, dest, save_ti);
 		break;
 	case PLAYLIST_VIEW:
-		if (worker_has_job(JOB_TYPE_PL))
-			goto worker_running;
-		dest = extended ? &pl_ext_filename : &pl_filename;
-		do_save(pl_for_each, arg, dest, save_ti);
+		pl_save();
 		break;
 	case QUEUE_VIEW:
-		if (worker_has_job(JOB_TYPE_QUEUE))
+		if (worker_has_job_by_type(JOB_TYPE_QUEUE))
 			goto worker_running;
 		dest = extended ? &play_queue_ext_filename : &play_queue_filename;
 		do_save(play_queue_for_each, arg, dest, save_ti);
 		break;
 	default:
-		info_msg(":save only works in views 1 & 2 (library) and 3 (playlist)");
+		info_msg(":save only works in views 1 - 4");
 	}
 	free(arg);
 	return;
@@ -353,11 +336,11 @@ struct window *current_win(void)
 	case TREE_VIEW:
 		return lib_cur_win;
 	case SORTED_VIEW:
-		return lib_editable.win;
+		return lib_editable.shared->win;
 	case PLAYLIST_VIEW:
-		return pl_editable.win;
+		return pl_cursor_win();
 	case QUEUE_VIEW:
-		return pq_editable.win;
+		return pq_editable.shared->win;
 	case BROWSER_VIEW:
 		return browser_win;
 	case HELP_VIEW:
@@ -443,6 +426,8 @@ static void cmd_set(char *arg)
 		if (cur_view == TREE_VIEW) {
 			lib_track_win->changed = 1;
 			lib_tree_win->changed = 1;
+		} else if (cur_view == PLAYLIST_VIEW) {
+			pl_mark_for_redraw();
 		} else {
 			current_win()->changed = 1;
 		}
@@ -459,7 +444,7 @@ static void cmd_set(char *arg)
 
 		opt = option_find(arg);
 		if (opt) {
-			opt->get(opt->id, buf);
+			opt->get(opt->data, buf, OPTION_MAX_SIZE);
 			info_msg("setting: '%s=%s'", arg, buf);
 		}
 	}
@@ -476,11 +461,13 @@ static void cmd_toggle(char *arg)
 		error_msg("%s is not toggle option", opt->name);
 		return;
 	}
-	opt->toggle(opt->id);
+	opt->toggle(opt->data);
 	help_win->changed = 1;
 	if (cur_view == TREE_VIEW) {
 		lib_track_win->changed = 1;
 		lib_tree_win->changed = 1;
+	} else if (cur_view == PLAYLIST_VIEW) {
+		pl_mark_for_redraw();
 	} else {
 		current_win()->changed = 1;
 	}
@@ -552,7 +539,7 @@ inside:
 	}
 
 	if (!*arg) {
-		player_seek(seek, relative, 1);
+		player_seek(seek, relative, 0);
 		return;
 	}
 err:
@@ -561,23 +548,17 @@ err:
 
 static void cmd_factivate(char *arg)
 {
-	editable_lock();
 	filters_activate_names(arg);
-	editable_unlock();
 }
 
 static void cmd_live_filter(char *arg)
 {
-	editable_lock();
 	filters_set_live(arg);
-	editable_unlock();
 }
 
 static void cmd_filter(char *arg)
 {
-	editable_lock();
 	filters_set_anonymous(arg);
-	editable_unlock();
 }
 
 static void cmd_fset(char *arg)
@@ -592,13 +573,12 @@ static void cmd_help(char *arg)
 
 static void cmd_invert(char *arg)
 {
-	editable_lock();
 	switch (cur_view) {
 	case SORTED_VIEW:
 		editable_invert_marks(&lib_editable);
 		break;
 	case PLAYLIST_VIEW:
-		editable_invert_marks(&pl_editable);
+		pl_invert_marks();
 		break;
 	case QUEUE_VIEW:
 		editable_invert_marks(&pq_editable);
@@ -606,18 +586,16 @@ static void cmd_invert(char *arg)
 	default:
 		info_msg(":invert only works in views 2-4");
 	}
-	editable_unlock();
 }
 
 static void cmd_mark(char *arg)
 {
-	editable_lock();
 	switch (cur_view) {
 	case SORTED_VIEW:
 		editable_mark(&lib_editable, arg);
 		break;
 	case PLAYLIST_VIEW:
-		editable_mark(&pl_editable, arg);
+		pl_mark(arg);
 		break;
 	case QUEUE_VIEW:
 		editable_mark(&pq_editable, arg);
@@ -625,18 +603,16 @@ static void cmd_mark(char *arg)
 	default:
 		info_msg(":mark only works in views 2-4");
 	}
-	editable_unlock();
 }
 
 static void cmd_unmark(char *arg)
 {
-	editable_lock();
 	switch (cur_view) {
 	case SORTED_VIEW:
 		editable_unmark(&lib_editable);
 		break;
 	case PLAYLIST_VIEW:
-		editable_unmark(&pl_editable);
+		pl_unmark();
 		break;
 	case QUEUE_VIEW:
 		editable_unmark(&pq_editable);
@@ -644,7 +620,6 @@ static void cmd_unmark(char *arg)
 	default:
 		info_msg(":unmark only works in views 2-4");
 	}
-	editable_unlock();
 }
 
 static void cmd_update_cache(char *arg)
@@ -731,7 +706,7 @@ static void cmd_unbind(char *arg)
 	if (*key == 0)
 		goto err;
 
-	/* FIXME: remove spaces at end */
+	strip_trailing_spaces(key);
 
 	key_unbind(arg, key, flag == 'f');
 	return;
@@ -752,7 +727,7 @@ static void cmd_showbind(char *arg)
 	if (*key == 0)
 		goto err;
 
-	/* FIXME: remove spaces at end */
+	strip_trailing_spaces(key);
 
 	show_binding(arg, key);
 	return;
@@ -763,7 +738,7 @@ err:
 static void cmd_quit(char *arg)
 {
 	int flag = parse_flags((const char **)&arg, "i");
-	if (!worker_has_job(JOB_TYPE_ANY)) {
+	if (!worker_has_job_by_type(JOB_TYPE_ANY)) {
 		if (flag != 'i' || yes_no_query("Quit cmus? [y/N]"))
 			cmus_running = 0;
 	} else {
@@ -774,10 +749,8 @@ static void cmd_quit(char *arg)
 
 static void cmd_reshuffle(char *arg)
 {
-	editable_lock();
 	lib_reshuffle();
 	pl_reshuffle();
-	editable_unlock();
 }
 
 static void cmd_source(char *arg)
@@ -795,7 +768,7 @@ static void cmd_colorscheme(char *arg)
 
 	snprintf(filename, sizeof(filename), "%s/%s.theme", cmus_config_dir, arg);
 	if (source_file(filename) == -1) {
-		snprintf(filename, sizeof(filename), DATADIR "/cmus/%s.theme", arg);
+		snprintf(filename, sizeof(filename), "%s/%s.theme", cmus_data_dir, arg);
 		if (source_file(filename) == -1)
 			error_msg("sourcing %s: %s", filename, strerror(errno));
 	}
@@ -1021,7 +994,6 @@ static void cmd_run(char *arg)
 	}
 
 	/* collect selected files (struct track_info) */
-	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
 		_tree_for_each_sel(add_ti, &sel, 0);
@@ -1030,13 +1002,12 @@ static void cmd_run(char *arg)
 		_editable_for_each_sel(&lib_editable, add_ti, &sel, 0);
 		break;
 	case PLAYLIST_VIEW:
-		_editable_for_each_sel(&pl_editable, add_ti, &sel, 0);
+		_pl_for_each_sel(add_ti, &sel, 0);
 		break;
 	case QUEUE_VIEW:
 		_editable_for_each_sel(&pq_editable, add_ti, &sel, 0);
 		break;
 	}
-	editable_unlock();
 
 	if (sel.tis_nr == 0) {
 		/* no files selected, do nothing */
@@ -1159,7 +1130,6 @@ static void cmd_echo(char *arg)
 	/* get only the first selected track */
 	sel_ti = NULL;
 
-	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
 		_tree_for_each_sel(get_one_ti, &sel_ti, 0);
@@ -1168,13 +1138,12 @@ static void cmd_echo(char *arg)
 		_editable_for_each_sel(&lib_editable, get_one_ti, &sel_ti, 0);
 		break;
 	case PLAYLIST_VIEW:
-		_editable_for_each_sel(&pl_editable, get_one_ti, &sel_ti, 0);
+		_pl_for_each_sel(get_one_ti, &sel_ti, 0);
 		break;
 	case QUEUE_VIEW:
 		_editable_for_each_sel(&pq_editable, get_one_ti, &sel_ti, 0);
 		break;
 	}
-	editable_unlock();
 
 	if (sel_ti == NULL)
 		return;
@@ -1183,9 +1152,6 @@ static void cmd_echo(char *arg)
 	track_info_unref(sel_ti);
 }
 
-#define VF_RELATIVE	0x01
-#define VF_PERCENTAGE	0x02
-
 static int parse_vol_arg(const char *arg, int *value, unsigned int *flags)
 {
 	unsigned int f = 0;
@@ -1225,18 +1191,6 @@ err:
 	return -1;
 }
 
-static int calc_vol(int val, int old, int max_vol, unsigned int flags)
-{
-	if (flags & VF_RELATIVE) {
-		if (flags & VF_PERCENTAGE)
-			val = scale_from_percentage(val, max_vol);
-		val += old;
-	} else if (flags & VF_PERCENTAGE) {
-		val = scale_from_percentage(val, max_vol);
-	}
-	return clamp(val, 0, max_vol);
-}
-
 /*
  * :vol value [value]
  *
@@ -1261,22 +1215,13 @@ static void cmd_vol(char *arg)
 
 	free_str_array(values);
 
-	if (soft_vol) {
-		l = calc_vol(l, soft_vol_l, 100, lf);
-		r = calc_vol(r, soft_vol_r, 100, rf);
-		player_set_soft_volume(l, r);
+	int rc = player_set_vol(l, lf, r, rf);
+	if (rc != OP_ERROR_SUCCESS) {
+		char *msg = op_get_error_msg(rc, "can't change volume");
+		error_msg("%s", msg);
+		free(msg);
 	} else {
-		int rc;
-		mixer_read_volume();
-		l = calc_vol(l, volume_l, volume_max, lf);
-		r = calc_vol(r, volume_r, volume_max, rf);
-		rc = mixer_set_volume(l, r);
-		if (rc != OP_ERROR_SUCCESS) {
-			char *msg = op_get_error_msg(rc, "can't change volume");
-			error_msg("%s", msg);
-			free(msg);
-		}
-		mixer_read_volume();
+		mpris_volume_changed();
 	}
 	update_statusline();
 	return;
@@ -1287,20 +1232,90 @@ err:
 
 static void cmd_prev_view(char *arg)
 {
-	int tmp;
 	if (prev_view >= 0) {
-		tmp = cur_view;
 		set_view(prev_view);
-		prev_view = tmp;
 	}
 }
 
+static void cmd_left_view(char *arg)
+{
+	if (cur_view == TREE_VIEW) {
+		set_view(HELP_VIEW);
+	} else {
+		set_view(cur_view - 1);
+	}
+}
+
+static void cmd_right_view(char *arg)
+{
+	if (cur_view == HELP_VIEW) {
+		set_view(TREE_VIEW);
+	} else {
+		set_view(cur_view + 1);
+	}
+}
+
+static void cmd_pl_create(char *arg)
+{
+	pl_create(arg);
+}
+
+static void cmd_pl_export(char *arg)
+{
+	if (cur_view == PLAYLIST_VIEW)
+		pl_export_selected_pl(arg);
+	else
+		info_msg(":pl-export only works in view 3");
+}
+
+static char *get_browser_add_file(void)
+{
+	char *sel = browser_get_sel();
+
+	if (sel && (ends_with(sel, "/../") || ends_with(sel, "/.."))) {
+		info_msg("For convenience, you can not add \"..\" directory from the browser view");
+		free(sel);
+		sel = NULL;
+	}
+
+	return sel;
+}
+
+static void cmd_pl_import(char *arg)
+{
+	char *name = NULL;
+
+	if (arg)
+		name = view_load_prepare(arg);
+	else if (cur_view == BROWSER_VIEW)
+		name = get_browser_add_file();
+	else
+		error_msg("not enough arguments");
+
+	if (name) {
+		pl_import(name);
+		free(name);
+	}
+}
+
+static void cmd_pl_rename(char *arg)
+{
+	if (cur_view == PLAYLIST_VIEW)
+		pl_rename_selected_pl(arg);
+	else
+		info_msg(":pl-rename only works in view 3");
+}
+
+static void cmd_version(char *arg)
+{
+	info_msg(VERSION);
+}
+
 static void cmd_view(char *arg)
 {
 	int view;
 
 	if (parse_enum(arg, 1, NR_VIEWS, view_names, &view) && (view - 1) != cur_view) {
-		prev_view = cur_view;
 		set_view(view - 1);
 	}
 }
@@ -1321,10 +1336,17 @@ static void cmd_p_pause(char *arg)
 	player_pause();
 }
 
+static void cmd_p_pause_playback(char *arg)
+{
+	player_pause_playback();
+}
+
 static void cmd_p_play(char *arg)
 {
 	if (arg) {
-		cmus_play_file(arg);
+		char *tmp = expand_filename(arg);
+		cmus_play_file(tmp);
+		free(tmp);
 	} else {
 		player_play();
 	}
@@ -1352,6 +1374,28 @@ static void cmd_pwd(char *arg)
 	}
 }
 
+static void cmd_raise_vte(char *arg)
+{
+	cmus_raise_vte();
+}
+
+static void cmd_rand(char *arg)
+{
+	switch (cur_view) {
+	case TREE_VIEW:
+		break;
+	case SORTED_VIEW:
+		editable_rand(&lib_editable);
+		break;
+	case PLAYLIST_VIEW:
+		pl_rand();
+		break;
+	case QUEUE_VIEW:
+		editable_rand(&pq_editable);
+		break;
+	}
+}
+
 static void cmd_search_next(char *arg)
 {
 	if (search_str) {
@@ -1368,17 +1412,22 @@ static void cmd_search_prev(char *arg)
 	}
 }
 
-static int sorted_for_each_sel(int (*cb)(void *data, struct track_info *ti), void *data, int reverse)
+static void cmd_search_start(char *arg)
+{
+	enter_search_mode();
+}
+
+static void cmd_search_b_start(char *arg)
+{
+	enter_search_backward_mode();
+}
+
+static int sorted_for_each_sel(track_info_cb cb, void *data, int reverse)
 {
 	return editable_for_each_sel(&lib_editable, cb, data, reverse);
 }
 
-static int pl_for_each_sel(int (*cb)(void *data, struct track_info *ti), void *data, int reverse)
-{
-	return editable_for_each_sel(&pl_editable, cb, data, reverse);
-}
-
-static int pq_for_each_sel(int (*cb)(void *data, struct track_info *ti), void *data, int reverse)
+static int pq_for_each_sel(track_info_cb cb, void *data, int reverse)
 {
 	return editable_for_each_sel(&pq_editable, cb, data, reverse);
 }
@@ -1400,27 +1449,21 @@ static int wrapper_cb(void *data, struct track_info *ti)
 {
 	struct wrapper_cb_data *add = data;
 
-	add->cb(ti);
+	add->cb(ti, NULL);
 	return 0;
 }
 
 static void add_from_browser(add_ti_cb add, int job_type)
 {
-	char *sel = browser_get_sel();
+	char *sel = get_browser_add_file();
 
 	if (sel) {
 		enum file_type ft;
 		char *ret;
 
-		if (ends_with(sel, "/../") || ends_with(sel, "/..")) {
-			info_msg("For convenience, you can not add \"..\" directory from the browser view");
-			free(sel);
-			return;
-		}
-
 		ft = cmus_detect_ft(sel, &ret);
 		if (ft != FILE_TYPE_INVALID) {
-			cmus_add(add, ret, ft, job_type, 0);
+			cmus_add(add, ret, ft, job_type, 0, NULL);
 			window_down(browser_win, 1);
 		}
 		free(ret);
@@ -1435,9 +1478,7 @@ static void cmd_win_add_l(char *arg)
 
 	if (cur_view <= QUEUE_VIEW) {
 		struct wrapper_cb_data add = { lib_add_track };
-		editable_lock();
 		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
-		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(lib_add_track, JOB_TYPE_LIB);
 	}
@@ -1445,17 +1486,19 @@ static void cmd_win_add_l(char *arg)
 
 static void cmd_win_add_p(char *arg)
 {
-	/* could allow adding dups? */
-	if (cur_view == PLAYLIST_VIEW)
+	if (cur_view == PLAYLIST_VIEW && pl_visible_is_marked())
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
-		struct wrapper_cb_data add = { pl_add_track };
-		editable_lock();
+		struct wrapper_cb_data add = { pl_add_track_to_marked_pl2 };
 		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
-		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
-		add_from_browser(pl_add_track, JOB_TYPE_PL);
+		char *sel = get_browser_add_file();
+		if (sel) {
+			if (pl_add_file_to_marked_pl(sel))
+				window_down(browser_win, 1);
+			free(sel);
+		}
 	}
 }
 
@@ -1466,9 +1509,7 @@ static void cmd_win_add_Q(char *arg)
 
 	if (cur_view <= QUEUE_VIEW) {
 		struct wrapper_cb_data add = { play_queue_prepend };
-		editable_lock();
 		view_for_each_sel[cur_view](wrapper_cb, &add, 1);
-		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(play_queue_prepend, JOB_TYPE_QUEUE);
 	}
@@ -1481,9 +1522,7 @@ static void cmd_win_add_q(char *arg)
 
 	if (cur_view <= QUEUE_VIEW) {
 		struct wrapper_cb_data add = { play_queue_append };
-		editable_lock();
 		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
-		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(play_queue_append, JOB_TYPE_QUEUE);
 	}
@@ -1499,12 +1538,8 @@ static void cmd_win_activate(char *arg)
 		if (lib_cur_track)
 			previous = &lib_cur_track->shuffle_track;
 		shuffle_root = &lib_shuffle_root;
-	} else if (cur_view == PLAYLIST_VIEW) {
-		previous = (struct shuffle_track *)pl_cur_track;
-		shuffle_root = &pl_shuffle_root;
 	}
 
-	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
 		info = tree_activate_selected();
@@ -1515,8 +1550,7 @@ static void cmd_win_activate(char *arg)
 		next = &lib_cur_track->shuffle_track;
 		break;
 	case PLAYLIST_VIEW:
-		info = pl_activate_selected();
-		next = (struct shuffle_track *)pl_cur_track;
+		info = pl_play_selected_row();
 		break;
 	case QUEUE_VIEW:
 		break;
@@ -1524,16 +1558,15 @@ static void cmd_win_activate(char *arg)
 		browser_enter();
 		break;
 	case FILTERS_VIEW:
-		filters_activate();
+		filters_activate(1);
 		break;
 	case HELP_VIEW:
 		help_select();
 		break;
 	}
-	editable_unlock();
 
 	if (info) {
-		if (shuffle)
+		if (shuffle && next)
 			shuffle_insert(shuffle_root, previous, next);
 		/* update lib/pl mode */
 		if (cur_view < 2)
@@ -1547,57 +1580,36 @@ static void cmd_win_activate(char *arg)
 
 static void cmd_win_mv_after(char *arg)
 {
-	editable_lock();
 	switch (cur_view) {
-	case TREE_VIEW:
-		break;
 	case SORTED_VIEW:
 		editable_move_after(&lib_editable);
 		break;
 	case PLAYLIST_VIEW:
-		editable_move_after(&pl_editable);
+		pl_win_mv_after();
 		break;
 	case QUEUE_VIEW:
 		editable_move_after(&pq_editable);
 		break;
-	case BROWSER_VIEW:
-		break;
-	case FILTERS_VIEW:
-		break;
-	case HELP_VIEW:
-		break;
 	}
-	editable_unlock();
 }
 
 static void cmd_win_mv_before(char *arg)
 {
-	editable_lock();
 	switch (cur_view) {
-	case TREE_VIEW:
-		break;
 	case SORTED_VIEW:
 		editable_move_before(&lib_editable);
 		break;
 	case PLAYLIST_VIEW:
-		editable_move_before(&pl_editable);
+		pl_win_mv_before();
 		break;
 	case QUEUE_VIEW:
 		editable_move_before(&pq_editable);
 		break;
-	case BROWSER_VIEW:
-		break;
-	case FILTERS_VIEW:
-		break;
-	case HELP_VIEW:
-		break;
 	}
-	editable_unlock();
 }
 
 static void cmd_win_remove(char *arg)
 {
-	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
 		tree_remove_sel();
@@ -1606,7 +1618,7 @@ static void cmd_win_remove(char *arg)
 		editable_remove_sel(&lib_editable);
 		break;
 	case PLAYLIST_VIEW:
-		editable_remove_sel(&pl_editable);
+		pl_win_remove();
 		break;
 	case QUEUE_VIEW:
 		editable_remove_sel(&pq_editable);
@@ -1621,12 +1633,10 @@ static void cmd_win_remove(char *arg)
 		help_remove();
 		break;
 	}
-	editable_unlock();
 }
 
 static void cmd_win_sel_cur(char *arg)
 {
-	editable_lock();
 	switch (cur_view) {
 	case TREE_VIEW:
 		tree_sel_current(auto_expand_albums_selcur);
@@ -1635,44 +1645,25 @@ static void cmd_win_sel_cur(char *arg)
 		sorted_sel_current();
 		break;
 	case PLAYLIST_VIEW:
-		pl_sel_current();
-		break;
-	case QUEUE_VIEW:
-		break;
-	case BROWSER_VIEW:
-		break;
-	case FILTERS_VIEW:
-		break;
-	case HELP_VIEW:
+		pl_select_playing_track();
 		break;
 	}
-	editable_unlock();
 }
 
 static void cmd_win_toggle(char *arg)
 {
 	switch (cur_view) {
 	case TREE_VIEW:
-		editable_lock();
 		tree_toggle_expand_artist();
-		editable_unlock();
 		break;
 	case SORTED_VIEW:
-		editable_lock();
 		editable_toggle_mark(&lib_editable);
-		editable_unlock();
 		break;
 	case PLAYLIST_VIEW:
-		editable_lock();
-		editable_toggle_mark(&pl_editable);
-		editable_unlock();
+		pl_win_toggle();
 		break;
 	case QUEUE_VIEW:
-		editable_lock();
 		editable_toggle_mark(&pq_editable);
-		editable_unlock();
-		break;
-	case BROWSER_VIEW:
 		break;
 	case FILTERS_VIEW:
 		filters_toggle_filter();
@@ -1685,23 +1676,17 @@ static void cmd_win_toggle(char *arg)
 
 static void cmd_win_scroll_down(char *arg)
 {
-	editable_lock();
 	window_scroll_down(current_win());
-	editable_unlock();
 }
 
 static void cmd_win_scroll_up(char *arg)
 {
-	editable_lock();
 	window_scroll_up(current_win());
-	editable_unlock();
 }
 
 static void cmd_win_bottom(char *arg)
 {
-	editable_lock();
 	window_goto_bottom(current_win());
-	editable_unlock();
 }
 
 static void cmd_win_down(char *arg)
@@ -1716,53 +1701,50 @@ static void cmd_win_down(char *arg)
 		}
 	}
 
-	editable_lock();
 	window_down(current_win(), num_rows);
-	editable_unlock();
 }
 
 static void cmd_win_next(char *arg)
 {
-	if (cur_view == TREE_VIEW) {
-		editable_lock();
+	if (cur_view == TREE_VIEW)
 		tree_toggle_active_window();
-		editable_unlock();
-	}
+	else if (cur_view == PLAYLIST_VIEW)
+		pl_win_next();
 }
 
 static void cmd_win_pg_down(char *arg)
 {
-	editable_lock();
 	window_page_down(current_win());
-	editable_unlock();
 }
 
 static void cmd_win_pg_up(char *arg)
 {
-	editable_lock();
 	window_page_up(current_win());
-	editable_unlock();
+}
+
+static void cmd_win_hf_pg_down(char *arg)
+{
+	window_half_page_down(current_win());
+}
+
+static void cmd_win_hf_pg_up(char *arg)
+{
+	window_half_page_up(current_win());
 }
 
 static void cmd_win_pg_top(char *arg)
 {
-	editable_lock();
 	window_page_top(current_win());
-	editable_unlock();
 }
 
 static void cmd_win_pg_bottom(char *arg)
 {
-	editable_lock();
 	window_page_bottom(current_win());
-	editable_unlock();
 }
 
 static void cmd_win_pg_middle(char *arg)
 {
-	editable_lock();
 	window_page_middle(current_win());
-	editable_unlock();
 }
 
 static void cmd_win_update_cache(char *arg)
@@ -1773,9 +1755,7 @@ static void cmd_win_update_cache(char *arg)
 	if (cur_view != TREE_VIEW && cur_view != SORTED_VIEW)
 		return;
 
-	editable_lock();
 	view_for_each_sel[cur_view](add_ti, &sel, 0);
-	editable_unlock();
 	if (sel.tis_nr == 0)
 		return;
 	sel.tis[sel.tis_nr] = NULL;
@@ -1784,9 +1764,7 @@ static void cmd_win_update_cache(char *arg)
 
 static void cmd_win_top(char *arg)
 {
-	editable_lock();
 	window_goto_top(current_win());
-	editable_unlock();
 }
 
 static void cmd_win_up(char *arg)
@@ -1801,9 +1779,7 @@ static void cmd_win_up(char *arg)
 		}
 	}
 
-	editable_lock();
 	window_up(current_win(), num_rows);
-	editable_unlock();
 }
 
 static void cmd_win_update(char *arg)
@@ -1814,10 +1790,7 @@ static void cmd_win_update(char *arg)
 		cmus_update_lib();
 		break;
 	case PLAYLIST_VIEW:
-		editable_lock();
-		editable_clear(&pl_editable);
-		editable_unlock();
-		cmus_add(pl_add_track, pl_filename, FILE_TYPE_PL, JOB_TYPE_PL, 0);
+		pl_win_update();
 		break;
 	case BROWSER_VIEW:
 		browser_reload();
@@ -1935,12 +1908,11 @@ static void cmd_lqueue(char *arg)
 		}
 		count = val;
 	}
-	editable_lock();
 	nmax = count_albums();
 	if (count > nmax)
 		count = nmax;
 	if (!count)
-		goto unlock;
+		return;
 
 	r = rand_array(count, nmax);
 	album = to_album(rb_first(&to_artist(rb_first(&lib_artist_root))->album_root));
@@ -1972,12 +1944,10 @@ static void cmd_lqueue(char *arg)
 		struct rb_node *tmp;
 
 		rb_for_each_entry(t, tmp, &a->album->track_root, tree_node)
-			play_queue_append(tree_track_info(t));
+			play_queue_append(tree_track_info(t), NULL);
 		free(a);
 		item = next;
 	} while (item != &head);
-unlock:
-	editable_unlock();
 }
 
 struct track_list {
@@ -2001,11 +1971,10 @@ static void cmd_tqueue(char *arg)
 		}
 		count = val;
 	}
-	editable_lock();
 	if (count > lib_editable.nr_tracks)
 		count = lib_editable.nr_tracks;
 	if (!count)
-		goto unlock;
+		return;
 
 	r = rand_array(count, lib_editable.nr_tracks);
 	item = lib_editable.head.next;
@@ -2027,12 +1996,10 @@ static void cmd_tqueue(char *arg)
 	do {
 		struct list_head *next = item->next;
 		struct track_list *t = container_of(item, struct track_list, node);
-		play_queue_append(t->track->info);
+		play_queue_append(t->track->info, NULL);
 		free(t);
 		item = next;
 	} while (item != &head);
-unlock:
-	editable_unlock();
 }
 
 /* tab exp {{{
@@ -2443,7 +2410,7 @@ static void expand_options(const char *str)
 					tails = xnew(char *, 1);
 
 					buf[0] = 0;
-					opt->get(opt->id, buf);
+					opt->get(opt->data, buf, OPTION_MAX_SIZE);
 					tails[0] = xstrdup(buf);
 
 					tabexp.head = xstrdup(str);
@@ -2541,7 +2508,7 @@ static void expand_colorscheme(const char *str)
 	PTR_ARRAY(array);
 
 	load_themes(cmus_config_dir, str, &array);
-	load_themes(DATADIR "/cmus", str, &array);
+	load_themes(cmus_data_dir, str, &array);
 
 	if (array.count) {
 		ptr_array_sort(&array, strptrcmp);
@@ -2558,76 +2525,90 @@ static void expand_commands(const char *str);
 
 /* sort by name */
 struct command commands[] = {
-	{ "add",		cmd_add,	1, 1, expand_add,	  0, 0 },
-	{ "bind",		cmd_bind,	1, 1, expand_bind_args,	  0, CMD_UNSAFE },
-	{ "browser-up",		cmd_browser_up,	0, 0, NULL,		  0, 0 },
-	{ "cd",			cmd_cd,		0, 1, expand_directories, 0, 0 },
-	{ "clear",		cmd_clear,	0, 1, NULL,		  0, 0 },
-	{ "colorscheme",	cmd_colorscheme,1, 1, expand_colorscheme, 0, 0 },
-	{ "echo",		cmd_echo,	1,-1, NULL,		  0, 0 },
-	{ "factivate",		cmd_factivate,	0, 1, expand_factivate,	  0, 0 },
-	{ "filter",		cmd_filter,	0, 1, NULL,		  0, 0 },
-	{ "fset",		cmd_fset,	1, 1, expand_fset,	  0, 0 },
-	{ "help",		cmd_help,	0, 0, NULL,		  0, 0 },
-	{ "invert",		cmd_invert,	0, 0, NULL,		  0, 0 },
-	{ "live-filter",	cmd_live_filter,0, 1, NULL,		  0, CMD_LIVE },
-	{ "load",		cmd_load,	1, 1, expand_load_save,	  0, 0 },
-	{ "lqueue",		cmd_lqueue,	0, 1, NULL,		  0, 0 },
-	{ "mark",		cmd_mark,	0, 1, NULL,		  0, 0 },
-	{ "player-next",	cmd_p_next,	0, 0, NULL,		  0, 0 },
-	{ "player-pause",	cmd_p_pause,	0, 0, NULL,		  0, 0 },
-	{ "player-play",	cmd_p_play,	0, 1, expand_playable,	  0, 0 },
-	{ "player-prev",	cmd_p_prev,	0, 0, NULL,		  0, 0 },
-	{ "player-stop",	cmd_p_stop,	0, 0, NULL,		  0, 0 },
-	{ "prev-view",		cmd_prev_view,	0, 0, NULL,		  0, 0 },
-	{ "push",		cmd_push,	1,-1, expand_commands,	  0, 0 },
-	{ "pwd",		cmd_pwd,	0, 0, NULL,		  0, 0 },
-	{ "quit",		cmd_quit,	0, 1, NULL,		  0, 0 },
-	{ "refresh",		cmd_refresh,	0, 0, NULL,		  0, 0 },
-	{ "run",		cmd_run,	1,-1, expand_program_paths, 0, CMD_UNSAFE },
-	{ "save",		cmd_save,	0, 1, expand_load_save,	  0, CMD_UNSAFE },
-	{ "search-next",	cmd_search_next,0, 0, NULL,		  0, 0 },
-	{ "search-prev",	cmd_search_prev,0, 0, NULL,		  0, 0 },
-	{ "seek",		cmd_seek,	1, 1, NULL,		  0, 0 },
-	{ "set",		cmd_set,	1, 1, expand_options,	  0, 0 },
-	{ "shell",		cmd_shell,	1,-1, expand_program_paths, 0, CMD_UNSAFE },
-	{ "showbind",		cmd_showbind,	1, 1, expand_unbind_args, 0, 0 },
-	{ "shuffle",		cmd_reshuffle,	0, 0, NULL,		  0, 0 },
-	{ "source",		cmd_source,	1, 1, expand_files,	  0, CMD_UNSAFE },
-	{ "toggle",		cmd_toggle,	1, 1, expand_toptions,	  0, 0 },
-	{ "tqueue",		cmd_tqueue,	0, 1, NULL,		  0, 0 },
-	{ "unbind",		cmd_unbind,	1, 1, expand_unbind_args, 0, 0 },
-	{ "unmark",		cmd_unmark,	0, 0, NULL,		  0, 0 },
-	{ "update-cache",	cmd_update_cache,0, 1, NULL,		  0, 0 },
-	{ "view",		cmd_view,	1, 1, NULL,		  0, 0 },
-	{ "vol",		cmd_vol,	1, 2, NULL,		  0, 0 },
-	{ "w",			cmd_save,	0, 1, expand_load_save,	  0, CMD_UNSAFE },
-	{ "win-activate",	cmd_win_activate,0, 0, NULL,		  0, 0 },
-	{ "win-add-l",		cmd_win_add_l,	0, 0, NULL,		  0, 0 },
-	{ "win-add-p",		cmd_win_add_p,	0, 0, NULL,		  0, 0 },
-	{ "win-add-Q",		cmd_win_add_Q,	0, 0, NULL,		  0, 0 },
-	{ "win-add-q",		cmd_win_add_q,	0, 0, NULL,		  0, 0 },
-	{ "win-bottom",		cmd_win_bottom,	0, 0, NULL,		  0, 0 },
-	{ "win-down",		cmd_win_down,	0, 1, NULL,		  0, 0 },
-	{ "win-mv-after",	cmd_win_mv_after,0, 0, NULL,		  0, 0 },
-	{ "win-mv-before",	cmd_win_mv_before,0, 0, NULL,		  0, 0 },
-	{ "win-next",		cmd_win_next,	0, 0, NULL,		  0, 0 },
-	{ "win-page-bottom",	cmd_win_pg_bottom,0, 0, NULL,		  0, 0 },
-	{ "win-page-down",	cmd_win_pg_down,0, 0, NULL,		  0, 0 },
-	{ "win-page-middle",	cmd_win_pg_middle,0, 0, NULL,		  0, 0 },
-	{ "win-page-top",	cmd_win_pg_top,	0, 0, NULL,		  0, 0 },
-	{ "win-page-up",	cmd_win_pg_up,	0, 0, NULL,		  0, 0 },
-	{ "win-remove",		cmd_win_remove,	0, 0, NULL,		  0, CMD_UNSAFE },
-	{ "win-scroll-down",	cmd_win_scroll_down,0, 0, NULL,		  0, 0 },
-	{ "win-scroll-up",	cmd_win_scroll_up,0, 0, NULL,		  0, 0 },
-	{ "win-sel-cur",	cmd_win_sel_cur,0, 0, NULL,		  0, 0 },
-	{ "win-toggle",		cmd_win_toggle,	0, 0, NULL,		  0, 0 },
-	{ "win-top",		cmd_win_top,	0, 0, NULL,		  0, 0 },
-	{ "win-up",		cmd_win_up,	0, 1, NULL,		  0, 0 },
-	{ "win-update",		cmd_win_update,	0, 0, NULL,		  0, 0 },
-	{ "win-update-cache",	cmd_win_update_cache,0, 1, NULL,	  0, 0 },
-	{ "wq",			cmd_quit,	0, 1, NULL,		  0, 0 },
-	{ NULL,			NULL,		0, 0, 0,		  0, 0 }
+	{ "add",                   cmd_add,              1, 1,  expand_add,           0, 0          },
+	{ "bind",                  cmd_bind,             1, 1,  expand_bind_args,     0, CMD_UNSAFE },
+	{ "browser-up",            cmd_browser_up,       0, 0,  NULL,                 0, 0          },
+	{ "cd",                    cmd_cd,               0, 1,  expand_directories,   0, 0          },
+	{ "clear",                 cmd_clear,            0, 1,  NULL,                 0, 0          },
+	{ "colorscheme",           cmd_colorscheme,      1, 1,  expand_colorscheme,   0, 0          },
+	{ "echo",                  cmd_echo,             1, -1, NULL,                 0, 0          },
+	{ "factivate",             cmd_factivate,        0, 1,  expand_factivate,     0, 0          },
+	{ "filter",                cmd_filter,           0, 1,  NULL,                 0, 0          },
+	{ "fset",                  cmd_fset,             1, 1,  expand_fset,          0, 0          },
+	{ "help",                  cmd_help,             0, 0,  NULL,                 0, 0          },
+	{ "invert",                cmd_invert,           0, 0,  NULL,                 0, 0          },
+	{ "live-filter",           cmd_live_filter,      0, 1,  NULL,                 0, CMD_LIVE   },
+	{ "load",                  cmd_load,             1, 1,  expand_load_save,     0, 0          },
+	{ "lqueue",                cmd_lqueue,           0, 1,  NULL,                 0, 0          },
+	{ "mark",                  cmd_mark,             0, 1,  NULL,                 0, 0          },
+	{ "player-next",           cmd_p_next,           0, 0,  NULL,                 0, 0          },
+	{ "player-pause",          cmd_p_pause,          0, 0,  NULL,                 0, 0          },
+	{ "player-pause-playback", cmd_p_pause_playback, 0, 0,  NULL,                 0, 0          },
+	{ "player-play",           cmd_p_play,           0, 1,  expand_playable,      0, 0          },
+	{ "player-prev",           cmd_p_prev,           0, 0,  NULL,                 0, 0          },
+	{ "player-stop",           cmd_p_stop,           0, 0,  NULL,                 0, 0          },
+	{ "prev-view",             cmd_prev_view,        0, 0,  NULL,                 0, 0          },
+	{ "left-view",             cmd_left_view,        0, 0,  NULL,                 0, 0          },
+	{ "right-view",            cmd_right_view,       0, 0,  NULL,                 0, 0          },
+	{ "pl-create",             cmd_pl_create,        1, -1, NULL,                 0, 0          },
+	{ "pl-export",             cmd_pl_export,        1, -1, NULL,                 0, 0          },
+	{ "pl-import",             cmd_pl_import,        0, -1, NULL,                 0, 0          },
+	{ "pl-rename",             cmd_pl_rename,        1, -1, NULL,                 0, 0          },
+	{ "push",                  cmd_push,             1, -1, expand_commands,      0, 0          },
+	{ "pwd",                   cmd_pwd,              0, 0,  NULL,                 0, 0          },
+	{ "raise-vte",             cmd_raise_vte,        0, 0,  NULL,                 0, 0          },
+	{ "rand",                  cmd_rand,             0, 0,  NULL,                 0, 0          },
+	{ "quit",                  cmd_quit,             0, 1,  NULL,                 0, 0          },
+	{ "refresh",               cmd_refresh,          0, 0,  NULL,                 0, 0          },
+	{ "run",                   cmd_run,              1, -1, expand_program_paths, 0, CMD_UNSAFE },
+	{ "save",                  cmd_save,             0, 1,  expand_load_save,     0, CMD_UNSAFE },
+	{ "search-b-start",        cmd_search_b_start,   0, 0,  NULL,                 0, 0          },
+	{ "search-next",           cmd_search_next,      0, 0,  NULL,                 0, 0          },
+	{ "search-prev",           cmd_search_prev,      0, 0,  NULL,                 0, 0          },
+	{ "search-start",          cmd_search_start,     0, 0,  NULL,                 0, 0          },
+	{ "seek",                  cmd_seek,             1, 1,  NULL,                 0, 0          },
+	{ "set",                   cmd_set,              1, 1,  expand_options,       0, 0          },
+	{ "shell",                 cmd_shell,            1, -1, expand_program_paths, 0, CMD_UNSAFE },
+	{ "showbind",              cmd_showbind,         1, 1,  expand_unbind_args,   0, 0          },
+	{ "shuffle",               cmd_reshuffle,        0, 0,  NULL,                 0, 0          },
+	{ "source",                cmd_source,           1, 1,  expand_files,         0, CMD_UNSAFE },
+	{ "toggle",                cmd_toggle,           1, 1,  expand_toptions,      0, 0          },
+	{ "tqueue",                cmd_tqueue,           0, 1,  NULL,                 0, 0          },
+	{ "unbind",                cmd_unbind,           1, 1,  expand_unbind_args,   0, 0          },
+	{ "unmark",                cmd_unmark,           0, 0,  NULL,                 0, 0          },
+	{ "update-cache",          cmd_update_cache,     0, 1,  NULL,                 0, 0          },
+	{ "version",               cmd_version,          0, 0,  NULL,                 0, 0          },
+	{ "view",                  cmd_view,             1, 1,  NULL,                 0, 0          },
+	{ "vol",                   cmd_vol,              1, 2,  NULL,                 0, 0          },
+	{ "w",                     cmd_save,             0, 1,  expand_load_save,     0, CMD_UNSAFE },
+	{ "win-activate",          cmd_win_activate,     0, 0,  NULL,                 0, 0          },
+	{ "win-add-l",             cmd_win_add_l,        0, 0,  NULL,                 0, 0          },
+	{ "win-add-p",             cmd_win_add_p,        0, 0,  NULL,                 0, 0          },
+	{ "win-add-Q",             cmd_win_add_Q,        0, 0,  NULL,                 0, 0          },
+	{ "win-add-q",             cmd_win_add_q,        0, 0,  NULL,                 0, 0          },
+	{ "win-bottom",            cmd_win_bottom,       0, 0,  NULL,                 0, 0          },
+	{ "win-down",              cmd_win_down,         0, 1,  NULL,                 0, 0          },
+	{ "win-half-page-down",    cmd_win_hf_pg_down,   0, 0,  NULL,                 0, 0          },
+	{ "win-half-page-up",      cmd_win_hf_pg_up,     0, 0,  NULL,                 0, 0          },
+	{ "win-mv-after",          cmd_win_mv_after,     0, 0,  NULL,                 0, 0          },
+	{ "win-mv-before",         cmd_win_mv_before,    0, 0,  NULL,                 0, 0          },
+	{ "win-next",              cmd_win_next,         0, 0,  NULL,                 0, 0          },
+	{ "win-page-bottom",       cmd_win_pg_bottom,    0, 0,  NULL,                 0, 0          },
+	{ "win-page-down",         cmd_win_pg_down,      0, 0,  NULL,                 0, 0          },
+	{ "win-page-middle",       cmd_win_pg_middle,    0, 0,  NULL,                 0, 0          },
+	{ "win-page-top",          cmd_win_pg_top,       0, 0,  NULL,                 0, 0          },
+	{ "win-page-up",           cmd_win_pg_up,        0, 0,  NULL,                 0, 0          },
+	{ "win-remove",            cmd_win_remove,       0, 0,  NULL,                 0, CMD_UNSAFE },
+	{ "win-scroll-down",       cmd_win_scroll_down,  0, 0,  NULL,                 0, 0          },
+	{ "win-scroll-up",         cmd_win_scroll_up,    0, 0,  NULL,                 0, 0          },
+	{ "win-sel-cur",           cmd_win_sel_cur,      0, 0,  NULL,                 0, 0          },
+	{ "win-toggle",            cmd_win_toggle,       0, 0,  NULL,                 0, 0          },
+	{ "win-top",               cmd_win_top,          0, 0,  NULL,                 0, 0          },
+	{ "win-up",                cmd_win_up,           0, 1,  NULL,                 0, 0          },
+	{ "win-update",            cmd_win_update,       0, 0,  NULL,                 0, 0          },
+	{ "win-update-cache",      cmd_win_update_cache, 0, 1,  NULL,                 0, 0          },
+	{ "wq",                    cmd_quit,             0, 1,  NULL,                 0, 0          },
+	{ NULL,                    NULL,                 0, 0,  0,                    0, 0          }
 };
 
 /* fills tabexp struct */
@@ -2942,6 +2923,12 @@ void command_mode_ch(uchar ch)
 		}
 		input_mode = NORMAL_MODE;
 		break;
+	case 0x10: // ^P
+		command_mode_key(KEY_UP);
+		return;
+	case 0xE: // ^N
+		command_mode_key(KEY_DOWN);
+		return;
 	case 0x0A:
 		if (cmdline.blen) {
 			run_command(cmdline.line);
