@@ -40,6 +40,10 @@ int track_is_compilation(const struct keyval *comments)
 	if (c && is_freeform_true(c))
 		return 1;
 
+	c = keyvals_get_val(comments, "partofacompilation");
+	if (c && is_freeform_true(c))
+		return 1;
+
 	aa = keyvals_get_val(comments, "albumartist");
 	if (aa && is_various_artists(aa))
 		return 1;
@@ -70,6 +74,10 @@ int track_is_va_compilation(const struct keyval *comments)
 	if (c && is_freeform_true(c))
 		return 1;
 
+	c = keyvals_get_val(comments, "partofacompilation");
+	if (c && is_freeform_true(c))
+		return 1;
+
 	return 0;
 }
 
@@ -176,7 +184,7 @@ int comments_get_date(const struct keyval *comments, const char *key)
 
 static const char *interesting[] = {
 	"artist", "album", "title", "tracknumber", "discnumber", "genre",
-	"date", "compilation", "albumartist", "artistsort", "albumartistsort",
+	"date", "compilation", "partofacompilation", "albumartist", "artistsort", "albumartistsort",
 	"albumsort",
 	"originaldate",
 	"replaygain_track_gain",
@@ -185,6 +193,7 @@ static const char *interesting[] = {
 	"replaygain_album_peak",
 	"musicbrainz_trackid",
 	"comment",
+	"bpm",
 	"arranger", "composer", "conductor", "lyricist", "performer",
 	"remixer", "label", "publisher", "work", "opus", "partnumber", "part",
 	"subtitle", "media",
@@ -198,6 +207,7 @@ static struct {
 	{ "album_artist", "albumartist" },
 	{ "album artist", "albumartist" },
 	{ "disc", "discnumber" },
+	{ "tempo", "bpm" },
 	{ "track", "tracknumber" },
 	{ "WM/Year", "date" },
 	{ "WM/ArtistSortOrder", "artistsort" },
