@@ -24,30 +24,45 @@
 #include "rbtree.h"
 #include "track.h"
 #include "locking.h"
+#include "cmus.h"
+
+struct editable;
+
+typedef void (*editable_free_track)(struct editable *e, struct list_head *head);
+
+struct editable_shared {
+	struct editable *owner;
+
+	struct window *win;
+	sort_key_t *sort_keys;
+	char sort_str[128];
+	editable_free_track free_track;
+	struct searchable *searchable;
+};
 
 struct editable {
-	struct window *win;
 	struct list_head head;
 	struct rb_root tree_root;
 	unsigned int nr_tracks;
 	unsigned int nr_marked;
 	unsigned int total_time;
-	sort_key_t *sort_keys;
-	char sort_str[128];
-	struct searchable *searchable;
-
-	void (*free_track)(struct list_head *item);
+	struct editable_shared *shared;
 };
 
-extern pthread_mutex_t editable_mutex;
+void editable_shared_init(struct editable_shared *shared,
+		editable_free_track free_track);
+void editable_shared_set_sort_keys(struct editable_shared *shared,
+		sort_key_t *keys);
 
-void editable_init(struct editable *e, void (*free_track)(struct list_head *item));
+void editable_init(struct editable *e, struct editable_shared *shared,
+		int take_ownership);
+void editable_take_ownership(struct editable *e);
 void editable_add(struct editable *e, struct simple_track *track);
 void editable_add_before(struct editable *e, struct simple_track *track);
 void editable_remove_track(struct editable *e, struct simple_track *track);
 void editable_remove_sel(struct editable *e);
 void editable_sort(struct editable *e);
-void editable_set_sort_keys(struct editable *e, sort_key_t *keys);
+void editable_rand(struct editable *e);
 void editable_toggle_mark(struct editable *e);
 void editable_move_after(struct editable *e);
 void editable_move_before(struct editable *e);
@@ -57,11 +72,14 @@ void editable_remove_matching_tracks(struct editable *e,
 void editable_mark(struct editable *e, const char *filter);
 void editable_unmark(struct editable *e);
 void editable_invert_marks(struct editable *e);
-int _editable_for_each_sel(struct editable *e, int (*cb)(void *data, struct track_info *ti),
-		void *data, int reverse);
-int editable_for_each_sel(struct editable *e, int (*cb)(void *data, struct track_info *ti),
-		void *data, int reverse);
+int _editable_for_each_sel(struct editable *e, track_info_cb cb, void *data,
+		int reverse);
+int editable_for_each_sel(struct editable *e, track_info_cb cb, void *data,
+		int reverse);
+int editable_for_each(struct editable *e, track_info_cb cb, void *data,
+		int reverse);
 void editable_update_track(struct editable *e, struct track_info *old, struct track_info *new);
+int editable_empty(struct editable *e);
 
 static inline void editable_track_to_iter(struct editable *e, struct simple_track *track, struct iter *iter)
 {
@@ -70,7 +88,4 @@ static inline void editable_track_to_iter(struct editable *e, struct simple_trac
 	iter->data2 = NULL;
 }
 
-#define editable_lock() cmus_mutex_lock(&editable_mutex)
-#define editable_unlock() cmus_mutex_unlock(&editable_mutex)
-
 #endif
