@@ -264,6 +264,7 @@ const char * const id3_key_names[NUM_ID3_KEYS] = {
 	"comment",
 	"musicbrainz_trackid",
 	"media",
+	"bpm",
 };
 
 static int utf16_is_lsurrogate(uchar uch)
@@ -571,6 +572,7 @@ static struct {
 	{ "TPUB", ID3_PUBLISHER }, // TPUB can be both publisher or label
 	{ "TIT3", ID3_SUBTITLE },
 	{ "TMED", ID3_MEDIA },
+	{ "TBPM", ID3_BPM},
 
 	/* obsolete frames (2.2.0) */
 	{ "TP1",  ID3_ARTIST },
@@ -585,6 +587,7 @@ static struct {
 	{ "TS2",  ID3_ALBUMARTISTSORT },
 	{ "TSA",  ID3_ALBUMSORT },
 	{ "TCP",  ID3_COMPILATION },
+	{ "TBP",  ID3_BPM },
 };
 
 static int frame_tab_index(const char *id)
@@ -726,8 +729,10 @@ static void decode_normal(struct id3tag *id3, const char *buf, int len, int enco
 		 * This must be TPE2 frame; ignore it if ID3_ALBUMARTIST is
 		 * already present
 		 */
-		if (id3->v2[key])
+		if (id3->v2[key]) {
+			free(out);
 			return;
+		}
 	} else if (key == ID3_PUBLISHER) {
 		 add_v2(id3, ID3_LABEL, strdup(out));
 	}
@@ -930,7 +935,7 @@ static void decode_rva2(struct id3tag *id3, const char *buf, int len)
 	/*
 	 * This crazy code comes from Mutagen
 	 */
-	peak_bytes = min(4, (peak_bits + 7) >> 3);
+	peak_bytes = min_i(4, (peak_bits + 7) >> 3);
 	peak_shift = ((8 - (peak_bits & 7)) & 7) + (4 - peak_bytes) * 8;
 
 	if (len < rva2_min_len + peak_bytes) {
