@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2006 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,11 +11,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "output.h"
@@ -26,10 +25,15 @@
 #include "list.h"
 #include "debug.h"
 #include "ui_curses.h"
-#include "config/libdir.h"
+#include "options.h"
+#include "xstrjoin.h"
+#include "misc.h"
 
 #include <string.h>
+#include <strings.h>
+#include <stdio.h>
 #include <stdlib.h>
+#include <stdbool.h>
 #include <sys/types.h>
 #include <dirent.h>
 #include <dlfcn.h>
@@ -41,8 +45,8 @@ struct output_plugin {
 
 	const struct output_plugin_ops *pcm_ops;
 	const struct mixer_plugin_ops *mixer_ops;
-	const char * const *pcm_options;
-	const char * const *mixer_options;
+	const struct output_plugin_opt *pcm_options;
+	const struct mixer_plugin_opt *mixer_options;
 	int priority;
 
 	unsigned int pcm_initialized : 1;
@@ -50,16 +54,14 @@ struct output_plugin {
 	unsigned int mixer_open : 1;
 };
 
-static const char * const plugin_dir = LIBDIR "/cmus/op";
+static const char *plugin_dir;
 static LIST_HEAD(op_head);
 static struct output_plugin *op = NULL;
 
-int soft_vol;
-int soft_vol_l = 100;
-int soft_vol_r = 100;
-
 /* volume is between 0 and volume_max */
 int volume_max = 0;
+int volume_l = -1;
+int volume_r = -1;
 
 static void add_plugin(struct output_plugin *plugin)
 {
@@ -82,17 +84,19 @@ void op_load_plugins(void)
 	DIR *dir;
 	struct dirent *d;
 
+	plugin_dir = xstrjoin(cmus_lib_dir, "/op");
 	dir = opendir(plugin_dir);
 	if (dir == NULL) {
 		error_msg("couldn't open directory `%s': %s", plugin_dir, strerror(errno));
 		return;
 	}
-	while ((d = readdir(dir)) != NULL) {
+	while ((d = (struct dirent *) readdir(dir)) != NULL) {
 		char filename[256];
 		struct output_plugin *plug;
 		void *so, *symptr;
 		char *ext;
-		const char *sym;
+		const unsigned *abi_version_ptr;
+		bool err = false;
 
 		if (d->d_name[0] == '.')
 			continue;
@@ -106,29 +110,33 @@ void op_load_plugins(void)
 
 		so = dlopen(filename, RTLD_NOW);
 		if (so == NULL) {
-			error_msg("%s", dlerror());
+			d_print("%s: %s\n", filename, dlerror());
 			continue;
 		}
 
 		plug = xnew(struct output_plugin, 1);
 
-		sym = "op_pcm_ops";
-		if (!(plug->pcm_ops = dlsym(so, sym)))
-			goto sym_err;
-
-		sym = "op_pcm_options";
-		if (!(plug->pcm_options = dlsym(so, sym)))
-			goto sym_err;
-
-		sym = "op_priority";
-		symptr = dlsym(so, sym);
-		if (symptr == NULL)
-			goto sym_err;
+		plug->pcm_ops = dlsym(so, "op_pcm_ops");
+		plug->pcm_options = dlsym(so, "op_pcm_options");
+		symptr = dlsym(so, "op_priority");
+		abi_version_ptr = dlsym(so, "op_abi_version");
+		if (!plug->pcm_ops || !plug->pcm_options || !symptr) {
+			error_msg("%s: missing symbol", filename);
+			err = true;
+		}
+		if (!abi_version_ptr || *abi_version_ptr != OP_ABI_VERSION) {
+			error_msg("%s: incompatible plugin version", filename);
+			err = true;
+		}
+		if (err) {
+			free(plug);
+			dlclose(so);
+			continue;
+		}
 		plug->priority = *(int *)symptr;
 
 		plug->mixer_ops = dlsym(so, "op_mixer_ops");
 		plug->mixer_options = dlsym(so, "op_mixer_options");
-
 		if (plug->mixer_ops == NULL || plug->mixer_options == NULL) {
 			plug->mixer_ops = NULL;
 			plug->mixer_options = NULL;
@@ -141,11 +149,6 @@ void op_load_plugins(void)
 		plug->mixer_open = 0;
 
 		add_plugin(plug);
-		continue;
-sym_err:
-		error_msg("%s: symbol %s not found", filename, sym);
-		free(plug);
-		dlclose(so);
 	}
 	closedir(dir);
 }
@@ -182,7 +185,7 @@ void op_exit_plugins(void)
 	}
 }
 
-static void close_mixer(void)
+void mixer_close(void)
 {
 	volume_max = 0;
 	if (op && op->mixer_open) {
@@ -192,13 +195,8 @@ static void close_mixer(void)
 	}
 }
 
-static void open_mixer(void)
+void mixer_open(void)
 {
-	if (soft_vol) {
-		volume_max = 100;
-		return;
-	}
-
 	if (op == NULL)
 		return;
 
@@ -209,6 +207,7 @@ static void open_mixer(void)
 		rc = op->mixer_ops->open(&volume_max);
 		if (rc == 0) {
 			op->mixer_open = 1;
+			mixer_read_volume();
 		} else {
 			volume_max = 0;
 		}
@@ -222,10 +221,7 @@ static int select_plugin(struct output_plugin *o)
 
 	if (!o->pcm_initialized)
 		return -OP_ERROR_NOT_INITIALIZED;
-
-	close_mixer();
 	op = o;
-	open_mixer();
 	return 0;
 }
 
@@ -244,20 +240,26 @@ int op_select_any(void)
 {
 	struct output_plugin *o;
 	int rc = -OP_ERROR_NO_PLUGIN;
+	sample_format_t sf = sf_channels(2) | sf_rate(44100) | sf_bits(16) | sf_signed(1);
 
 	list_for_each_entry(o, &op_head, node) {
 		rc = select_plugin(o);
-		if (rc == 0)
+		if (rc != 0)
+			continue;
+		rc = o->pcm_ops->open(sf, NULL);
+		if (rc == 0) {
+			o->pcm_ops->close();
 			break;
+		}
 	}
 	return rc;
 }
 
-int op_open(sample_format_t sf)
+int op_open(sample_format_t sf, const channel_position_t *channel_map)
 {
 	if (op == NULL)
 		return -OP_ERROR_NOT_INITIALIZED;
-	return op->pcm_ops->open(sf);
+	return op->pcm_ops->open(sf, channel_map);
 }
 
 int op_drop(void)
@@ -274,10 +276,7 @@ int op_close(void)
 
 int op_write(const char *buffer, int count)
 {
-	int rc;
-
-	rc = op->pcm_ops->write(buffer, count);
-	return rc;
+	return op->pcm_ops->write(buffer, count);
 }
 
 int op_pause(void)
@@ -296,128 +295,132 @@ int op_unpause(void)
 
 int op_buffer_space(void)
 {
-	int rc;
-	
-	rc = op->pcm_ops->buffer_space();
-	return rc;
+	return op->pcm_ops->buffer_space();
 }
 
-int op_set_volume(int left, int right)
+int mixer_set_volume(int left, int right)
 {
-	if (soft_vol) {
-		soft_vol_l = left;
-		soft_vol_r = right;
-		return 0;
-	}
 	if (op == NULL)
 		return -OP_ERROR_NOT_INITIALIZED;
 	if (!op->mixer_open)
-		return -OP_ERROR_NOT_SUPPORTED;
+		return -OP_ERROR_NOT_OPEN;
 	return op->mixer_ops->set_volume(left, right);
 }
 
-int op_get_volume(int *left, int *right)
+int mixer_read_volume(void)
 {
-	if (soft_vol) {
-		*left = soft_vol_l;
-		*right = soft_vol_r;
-		return 0;
-	}
 	if (op == NULL)
 		return -OP_ERROR_NOT_INITIALIZED;
 	if (!op->mixer_open)
+		return -OP_ERROR_NOT_OPEN;
+	return op->mixer_ops->get_volume(&volume_l, &volume_r);
+}
+
+int mixer_get_fds(int *fds)
+{
+	if (op == NULL)
+		return -OP_ERROR_NOT_INITIALIZED;
+	if (!op->mixer_open)
+		return -OP_ERROR_NOT_OPEN;
+	if (!op->mixer_ops->get_fds)
 		return -OP_ERROR_NOT_SUPPORTED;
-	return op->mixer_ops->get_volume(left, right);
+	return op->mixer_ops->get_fds(fds);
 }
 
-void op_set_soft_vol(int soft)
+extern int soft_vol;
+
+static void option_error(int rc)
 {
-	close_mixer();
-	soft_vol = soft;
-	open_mixer();
+	char *msg = op_get_error_msg(rc, "setting option");
+	error_msg("%s", msg);
+	free(msg);
 }
 
-#define OP_OPT_ID(plugin_idx, is_mixer, option_idx) \
-	(((plugin_idx) << 16) | ((is_mixer) << 15) | (option_idx))
+static void set_dsp_option(void *data, const char *val)
+{
+	const struct output_plugin_opt *o = data;
+	int rc;
 
-static struct output_plugin *find_plugin(int idx)
+	rc = o->set(val);
+	if (rc)
+		option_error(rc);
+}
+
+static bool option_of_current_mixer(const struct mixer_plugin_opt *opt)
+{
+	const struct mixer_plugin_opt *mpo;
+
+	if (!op)
+		return false;
+	for (mpo = op->mixer_options; mpo && mpo->name; mpo++) {
+		if (mpo == opt)
+			return true;
+	}
+	return false;
+}
+
+static void set_mixer_option(void *data, const char *val)
+{
+	const struct mixer_plugin_opt *o = data;
+	int rc;
+
+	rc = o->set(val);
+	if (rc) {
+		option_error(rc);
+	} else if (option_of_current_mixer(o)) {
+		/* option of the current op was set
+		 * try to reopen the mixer */
+		mixer_close();
+		if (!soft_vol)
+			mixer_open();
+	}
+}
+
+static void get_dsp_option(void *data, char *buf, size_t size)
+{
+	const struct output_plugin_opt *o = data;
+	char *val = NULL;
+
+	o->get(&val);
+	if (val) {
+		strscpy(buf, val, size);
+		free(val);
+	}
+}
+
+static void get_mixer_option(void *data, char *buf, size_t size)
+{
+	const struct mixer_plugin_opt *o = data;
+	char *val = NULL;
+
+	o->get(&val);
+	if (val) {
+		strscpy(buf, val, size);
+		free(val);
+	}
+}
+
+void op_add_options(void)
 {
 	struct output_plugin *o;
-
-	list_for_each_entry(o, &op_head, node) {
-		if (idx == 0)
-			return o;
-		idx--;
-	}
-	return NULL;
-}
-
-int op_set_option(unsigned int id, const char *val)
-{
-	unsigned int pid = id >> 16;
-	unsigned int mix = id & 0x8000;
-	unsigned int oid = id & 0x7fff;
-	const struct output_plugin *o = find_plugin(pid);
-
-	if (o == NULL)
-		return -OP_ERROR_NOT_OPTION;
-
-	if (mix) {
-		const struct mixer_plugin_ops *mo = o->mixer_ops;
-		int rc = mo->set_option(oid, val);
-
-		if (rc == 0 && op && op->mixer_ops == mo) {
-			/* option of the current op was set
-			 * try to reopen the mixer */
-			close_mixer();
-			open_mixer();
-		}
-		return rc;
-	}
-	return o->pcm_ops->set_option(oid, val);
-}
-
-int op_get_option(unsigned int id, char **val)
-{
-	unsigned int pid = id >> 16;
-	unsigned int mix = id & 0x8000;
-	unsigned int oid = id & 0x7fff;
-	const struct output_plugin *o = find_plugin(pid);
-
-	if (o == NULL)
-		return -OP_ERROR_NOT_OPTION;
-
-	if (mix)
-		return o->mixer_ops->get_option(oid, val);
-	return o->pcm_ops->get_option(oid, val);
-}
-
-int op_for_each_option(void (*cb)(unsigned int id, const char *key))
-{
-	struct output_plugin *o;
-	unsigned int pid, oid;
+	const struct output_plugin_opt *opo;
+	const struct mixer_plugin_opt *mpo;
 	char key[64];
 
-	pid = -1;
 	list_for_each_entry(o, &op_head, node) {
-		pid++;
-
-		for (oid = 0; o->pcm_options[oid]; oid++) {
-			snprintf(key, sizeof(key), "dsp.%s.%s",
-					o->name,
-					o->pcm_options[oid]);
-			cb(OP_OPT_ID(pid, 0, oid), key);
+		for (opo = o->pcm_options; opo->name; opo++) {
+			snprintf(key, sizeof(key), "dsp.%s.%s", o->name,
+					opo->name);
+			option_add(xstrdup(key), opo, get_dsp_option,
+					set_dsp_option, NULL, 0);
 		}
-		if (o->mixer_ops == NULL)
-			continue;
-		for (oid = 0; o->mixer_options[oid]; oid++) {
-			snprintf(key, sizeof(key), "mixer.%s.%s",
-					o->name,
-					o->mixer_options[oid]);
-			cb(OP_OPT_ID(pid, 1, oid), key);
+		for (mpo = o->mixer_options; mpo && mpo->name; mpo++) {
+			snprintf(key, sizeof(key), "mixer.%s.%s", o->name,
+					mpo->name);
+			option_add(xstrdup(key), mpo, get_mixer_option,
+					set_mixer_option, NULL, 0);
 		}
 	}
-	return 0;
 }
 
 char *op_get_error_msg(int rc, const char *arg)
@@ -440,6 +443,10 @@ char *op_get_error_msg(int rc, const char *arg)
 		snprintf(buffer, sizeof(buffer),
 				"%s: function not supported", arg);
 		break;
+	case OP_ERROR_NOT_OPEN:
+		snprintf(buffer, sizeof(buffer),
+				"%s: mixer is not open", arg);
+		break;
 	case OP_ERROR_SAMPLE_FORMAT:
 		snprintf(buffer, sizeof(buffer),
 				"%s: sample format not supported", arg);
@@ -471,9 +478,9 @@ void op_dump_plugins(void)
 	}
 }
 
-char *op_get_current(void)
+const char *op_get_current(void)
 {
 	if (op)
-		return xstrdup(op->name);
+		return op->name;
 	return NULL;
 }
