@@ -1,6 +1,7 @@
-/* 
- * Copyright 2005 Timo Hirvonen
- * 
+/*
+ * Copyright 2008-2013 Various Authors
+ * Copyright 2004 Timo Hirvonen
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,17 +11,18 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
-#ifndef _PATH_H
-#define _PATH_H
+#ifndef CMUS_PATH_H
+#define CMUS_PATH_H
 
-extern void path_strip(char *str);
-extern char *path_absolute(const char *src);
+const char *get_extension(const char *filename);
+const char *path_basename(const char *path);
+void path_strip(char *str);
+char *path_absolute_cwd(const char *src, const char *cwd);
+char *path_absolute(const char *src);
 
 #endif
