@@ -21,11 +21,14 @@
 #include "input.h"
 #include "output.h"
 #include "sf.h"
+#include "op.h"
 #include "utils.h"
 #include "xmalloc.h"
 #include "debug.h"
 #include "compiler.h"
 #include "options.h"
+#include "mpris.h"
+#include "cmus.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -36,7 +39,9 @@
 #include <stdarg.h>
 #include <math.h>
 
-const char * const player_status_names[] = { "stopped", "playing", "paused", NULL };
+const char * const player_status_names[] = {
+	"stopped", "playing", "paused", NULL
+};
 
 enum producer_status {
 	PS_UNLOADED,
@@ -51,10 +56,12 @@ enum consumer_status {
 	CS_PAUSED
 };
 
-struct player_info player_info = {
-	.mutex = CMUS_MUTEX_INITIALIZER,
+/* protects player_info_priv and player_metadata */
+static pthread_mutex_t player_info_mutex = CMUS_MUTEX_INITIALIZER;
+struct player_info player_info;
+char player_metadata[255 * 16 + 1];
+static struct player_info player_info_priv = {
 	.ti = NULL,
-	.metadata = { 0, },
 	.status = PLAYER_STATUS_STOPPED,
 	.pos = 0,
 	.current_bitrate = -1,
@@ -82,8 +89,6 @@ int soft_vol;
 int soft_vol_l;
 int soft_vol_r;
 
-static const struct player_callbacks *player_cbs = NULL;
-
 static sample_format_t buffer_sf;
 static CHANNEL_MAP(buffer_channel_map);
 
@@ -109,6 +114,9 @@ static double replaygain_scale = 1.0;
 
 /* locking {{{ */
 
+#define player_info_priv_lock() cmus_mutex_lock(&player_info_mutex)
+#define player_info_priv_unlock() cmus_mutex_unlock(&player_info_mutex)
+
 #define producer_lock() cmus_mutex_lock(&producer_mutex)
 #define producer_unlock() cmus_mutex_unlock(&producer_mutex)
 
@@ -146,9 +154,7 @@ static void set_buffer_sf(void)
 	if (sf_get_channels(buffer_sf) <= 2 && sf_get_bits(buffer_sf) <= 16) {
 		buffer_sf &= SF_RATE_MASK;
 		buffer_sf |= sf_channels(2) | sf_bits(16) | sf_signed(1);
-#ifdef WORDS_BIGENDIAN
-		buffer_sf |= sf_bigendian(1);
-#endif
+		buffer_sf |= sf_host_endian();
 		channel_map_init_stereo(buffer_channel_map);
 	}
 }
@@ -342,24 +348,24 @@ static void update_rg_scale(void)
 	double gain, peak, db, scale, limit;
 
 	replaygain_scale = 1.0;
-	if (!player_info.ti || !replaygain)
+	if (!player_info_priv.ti || !replaygain)
 		return;
 
 	if (replaygain == RG_TRACK || replaygain == RG_TRACK_PREFERRED) {
-		gain = player_info.ti->rg_track_gain;
-		peak = player_info.ti->rg_track_peak;
+		gain = player_info_priv.ti->rg_track_gain;
+		peak = player_info_priv.ti->rg_track_peak;
 	} else {
-		gain = player_info.ti->rg_album_gain;
-		peak = player_info.ti->rg_album_peak;
+		gain = player_info_priv.ti->rg_album_gain;
+		peak = player_info_priv.ti->rg_album_peak;
 	}
 
 	if (isnan(gain)) {
 		if (replaygain == RG_TRACK_PREFERRED) {
-			gain = player_info.ti->rg_album_gain;
-			peak = player_info.ti->rg_album_peak;
+			gain = player_info_priv.ti->rg_album_gain;
+			peak = player_info_priv.ti->rg_album_peak;
 		} else if (replaygain == RG_ALBUM_PREFERRED) {
-			gain = player_info.ti->rg_track_gain;
-			peak = player_info.ti->rg_track_peak;
+			gain = player_info_priv.ti->rg_track_gain;
+			peak = player_info_priv.ti->rg_track_peak;
 		}
 	}
 
@@ -395,34 +401,29 @@ static inline unsigned int buffer_second_size(void)
 	return sf_get_second_size(buffer_sf);
 }
 
-static inline int get_next(struct track_info **ti)
-{
-	return player_cbs->get_next(ti);
-}
-
 /* updating player status {{{ */
 
 static inline void _file_changed(struct track_info *ti)
 {
-	if (player_info.ti)
-		track_info_unref(player_info.ti);
+	player_info_priv_lock();
+	if (player_info_priv.ti)
+		track_info_unref(player_info_priv.ti);
 
-	player_info.ti = ti;
+	player_info_priv.ti = ti;
 	update_rg_scale();
-	player_info.metadata[0] = 0;
-	player_info.file_changed = 1;
+	player_metadata[0] = 0;
+	player_info_priv.file_changed = 1;
+	player_info_priv_unlock();
 }
 
 static inline void file_changed(struct track_info *ti)
 {
-	player_info_lock();
 	if (ti) {
 		d_print("file: %s\n", ti->filename);
 	} else {
 		d_print("unloaded\n");
 	}
 	_file_changed(ti);
-	player_info_unlock();
 }
 
 static inline void metadata_changed(void)
@@ -430,37 +431,36 @@ static inline void metadata_changed(void)
 	struct keyval *comments;
 	int rc;
 
-	player_info_lock();
+	player_info_priv_lock();
 	if (ip_get_metadata(ip)) {
 		d_print("metadata changed: %s\n", ip_get_metadata(ip));
-		memcpy(player_info.metadata, ip_get_metadata(ip),
-		       255 * 16 + 1);
+		memcpy(player_metadata, ip_get_metadata(ip), 255 * 16 + 1);
 	}
 
 	rc = ip_read_comments(ip, &comments);
 	if (!rc) {
-		if (player_info.ti->comments)
-			keyvals_free(player_info.ti->comments);
-		track_info_set_comments(player_info.ti, comments);
+		if (player_info_priv.ti->comments)
+			keyvals_free(player_info_priv.ti->comments);
+		track_info_set_comments(player_info_priv.ti, comments);
 	}
 
-	player_info.metadata_changed = 1;
-	player_info_unlock();
+	player_info_priv.metadata_changed = 1;
+	player_info_priv_unlock();
 }
 
 static void player_error(const char *msg)
 {
-	player_info_lock();
-	player_info.status = (enum player_status)consumer_status;
-	player_info.pos = 0;
-	player_info.current_bitrate = -1;
-	player_info.buffer_fill = buffer_get_filled_chunks();
-	player_info.buffer_size = buffer_nr_chunks;
-	player_info.status_changed = 1;
+	player_info_priv_lock();
+	player_info_priv.status = (enum player_status)consumer_status;
+	player_info_priv.pos = 0;
+	player_info_priv.current_bitrate = -1;
+	player_info_priv.buffer_fill = buffer_get_filled_chunks();
+	player_info_priv.buffer_size = buffer_nr_chunks;
+	player_info_priv.status_changed = 1;
 
-	free(player_info.error_msg);
-	player_info.error_msg = xstrdup(msg);
-	player_info_unlock();
+	free(player_info_priv.error_msg);
+	player_info_priv.error_msg = xstrdup(msg);
+	player_info_priv_unlock();
 
 	d_print("ERROR: '%s'\n", msg);
 }
@@ -506,14 +506,14 @@ static void _producer_buffer_fill_update(void)
 {
 	int fill;
 
-	player_info_lock();
+	player_info_priv_lock();
 	fill = buffer_get_filled_chunks();
-	if (fill != player_info.buffer_fill) {
+	if (fill != player_info_priv.buffer_fill) {
 /* 		d_print("\n"); */
-		player_info.buffer_fill = fill;
-		player_info.buffer_fill_changed = 1;
+		player_info_priv.buffer_fill = fill;
+		player_info_priv.buffer_fill_changed = 1;
 	}
-	player_info_unlock();
+	player_info_priv_unlock();
 }
 
 /*
@@ -530,16 +530,16 @@ static void _consumer_position_update(void)
 	if (pos != old_pos) {
 		old_pos = pos;
 
-		player_info_lock();
-		player_info.pos = pos;
+		player_info_priv_lock();
+		player_info_priv.pos = pos;
 
 		if (show_current_bitrate) {
 			bitrate = ip_current_bitrate(ip);
 			if (bitrate != -1)
-				player_info.current_bitrate = bitrate;
+				player_info_priv.current_bitrate = bitrate;
 		}
-		player_info.position_changed = 1;
-		player_info_unlock();
+		player_info_priv.position_changed = 1;
+		player_info_priv_unlock();
 	}
 }
 
@@ -554,14 +554,14 @@ static void _player_status_changed(void)
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED)
 		pos = consumer_pos / buffer_second_size();
 
-	player_info_lock();
-	player_info.status = (enum player_status)consumer_status;
-	player_info.pos = pos;
-	player_info.current_bitrate = -1;
-	player_info.buffer_fill = buffer_get_filled_chunks();
-	player_info.buffer_size = buffer_nr_chunks;
-	player_info.status_changed = 1;
-	player_info_unlock();
+	player_info_priv_lock();
+	player_info_priv.status = (enum player_status)consumer_status;
+	player_info_priv.pos = pos;
+	player_info_priv.current_bitrate = -1;
+	player_info_priv.buffer_fill = buffer_get_filled_chunks();
+	player_info_priv.buffer_size = buffer_nr_chunks;
+	player_info_priv.status_changed = 1;
+	player_info_priv_unlock();
 }
 
 /* updating player status }}} */
@@ -631,7 +631,7 @@ static void _producer_play(void)
 	if (producer_status == PS_UNLOADED) {
 		struct track_info *ti;
 
-		if (get_next(&ti) == 0) {
+		if ((ti = cmus_get_next_track())) {
 			int rc;
 
 			ip = ip_new(ti->filename);
@@ -763,7 +763,6 @@ static void _consumer_pause(void)
 
 /* setting consumer status }}} */
 
-
 static int change_sf(int drop)
 {
 	int old_sf = buffer_sf;
@@ -804,8 +803,8 @@ static void _consumer_handle_eof(void)
 		return;
 	}
 
-	if (player_info.ti)
-		player_info.ti->play_count++;
+	if (player_info_priv.ti)
+		player_info_priv.ti->play_count++;
 
 	if (player_repeat_current) {
 		if (player_cont) {
@@ -814,12 +813,12 @@ static void _consumer_handle_eof(void)
 		} else {
 			_producer_stop();
 			_consumer_drain_and_stop();
-			_player_status_changed();
 		}
+		_player_status_changed();
 		return;
 	}
 
-	if (get_next(&ti) == 0) {
+	if ((ti = cmus_get_next_track())) {
 		_producer_unload();
 		ip = ip_new(ti->filename);
 		_producer_status_update(PS_STOPPED);
@@ -1012,7 +1011,7 @@ static void *producer_loop(void *arg)
 	return NULL;
 }
 
-void player_init(const struct player_callbacks *callbacks)
+void player_init(void)
 {
 	int rc;
 #ifdef REALTIME_SCHEDULING
@@ -1026,8 +1025,6 @@ void player_init(const struct player_callbacks *callbacks)
 	buffer_nr_chunks = 10 * 44100 * 16 / 8 * 2 / CHUNK_SIZE;
 	buffer_init();
 
-	player_cbs = callbacks;
-
 #ifdef REALTIME_SCHEDULING
 	rc = pthread_attr_init(&attr);
 	BUG_ON(rc);
@@ -1056,7 +1053,7 @@ void player_init(const struct player_callbacks *callbacks)
 	}
 	BUG_ON(rc);
 
-	/* update player_info.cont etc. */
+	/* update player_info_priv.cont etc. */
 	player_lock();
 	_player_status_changed();
 	player_unlock();
@@ -1143,6 +1140,12 @@ void player_pause(void)
 	player_unlock();
 }
 
+void player_pause_playback(void)
+{
+	if (consumer_status == CS_PLAYING)
+		player_pause();
+}
+
 void player_set_file(struct track_info *ti)
 {
 	player_lock();
@@ -1243,8 +1246,11 @@ void player_seek(double offset, int relative, int start_playing)
 				new_pos = 0.0;
 			if (offset > 0.0) {
 				/* seeking forward */
-				if (new_pos > duration - 5.0)
-					new_pos = duration - 5.0;
+				if (new_pos > duration) {
+					player_unlock();
+					cmus_next();
+					return;
+				}
 				if (new_pos < 0.0)
 					new_pos = 0.0;
 				if (new_pos < pos - 0.5) {
@@ -1286,6 +1292,7 @@ void player_seek(double offset, int relative, int start_playing)
 			d_print("error: ip_seek returned %d\n", rc);
 		}
 	}
+	mpris_seeked();
 	player_unlock();
 }
 
@@ -1378,6 +1385,35 @@ void player_set_soft_vol(int soft)
 	consumer_unlock();
 }
 
+static int calc_vol(int val, int old, int max_vol, unsigned int flags)
+{
+	if (flags & VF_RELATIVE) {
+		if (flags & VF_PERCENTAGE)
+			val = scale_from_percentage(val, max_vol);
+		val += old;
+	} else if (flags & VF_PERCENTAGE) {
+		val = scale_from_percentage(val, max_vol);
+	}
+	return clamp(val, 0, max_vol);
+}
+
+int player_set_vol(int l, int lf, int r, int rf)
+{
+	int rc = OP_ERROR_SUCCESS;
+	if (soft_vol) {
+		l = calc_vol(l, soft_vol_l, 100, lf);
+		r = calc_vol(r, soft_vol_r, 100, rf);
+		player_set_soft_volume(l, r);
+	} else {
+		mixer_read_volume();
+		l = calc_vol(l, volume_l, volume_max, lf);
+		r = calc_vol(r, volume_r, volume_max, rf);
+		rc = mixer_set_volume(l, r);
+		mixer_read_volume();
+	}
+	return rc;
+}
+
 void player_set_rg(enum replaygain rg)
 {
 	player_lock();
@@ -1386,9 +1422,9 @@ void player_set_rg(enum replaygain rg)
 		scale_pos = consumer_pos;
 	replaygain = rg;
 
-	player_info_lock();
+	player_info_priv_lock();
 	update_rg_scale();
-	player_info_unlock();
+	player_info_priv_unlock();
 
 	player_unlock();
 }
@@ -1398,9 +1434,9 @@ void player_set_rg_limit(int limit)
 	player_lock();
 	replaygain_limit = limit;
 
-	player_info_lock();
+	player_info_priv_lock();
 	update_rg_scale();
-	player_info_unlock();
+	player_info_priv_unlock();
 
 	player_unlock();
 }
@@ -1410,9 +1446,40 @@ void player_set_rg_preamp(double db)
 	player_lock();
 	replaygain_preamp = db;
 
-	player_info_lock();
+	player_info_priv_lock();
 	update_rg_scale();
-	player_info_unlock();
+	player_info_priv_unlock();
 
 	player_unlock();
 }
+
+void player_info_snapshot(void)
+{
+	player_info_priv_lock();
+
+	free(player_info.error_msg);
+	if (player_info.ti)
+		track_info_unref(player_info.ti);
+	memcpy(&player_info, &player_info_priv, sizeof(player_info));
+	if (player_info.ti)
+		track_info_ref(player_info.ti);
+
+	player_info_priv.file_changed = 0;
+	player_info_priv.metadata_changed = 0;
+	player_info_priv.status_changed = 0;
+	player_info_priv.position_changed = 0;
+	player_info_priv.buffer_fill_changed = 0;
+	player_info_priv.error_msg = NULL;
+
+	player_info_priv_unlock();
+}
+
+void player_metadata_lock(void)
+{
+	cmus_mutex_lock(&player_info_mutex);
+}
+
+void player_metadata_unlock(void)
+{
+	cmus_mutex_unlock(&player_info_mutex);
+}
