@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,18 +11,15 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
-#ifndef _PLAYER_H
-#define _PLAYER_H
+#ifndef CMUS_PLAYER_H
+#define CMUS_PLAYER_H
 
 #include "locking.h"
-#include "comment.h"
 #include "track_info.h"
 
 #include <pthread.h>
@@ -35,38 +33,30 @@ enum {
 	PLAYER_ERROR_NOT_SUPPORTED
 };
 
+extern const char * const player_status_names[];
 enum player_status {
 	PLAYER_STATUS_STOPPED,
 	PLAYER_STATUS_PLAYING,
-	PLAYER_STATUS_PAUSED
+	PLAYER_STATUS_PAUSED,
+	NR_PLAYER_STATUS
 };
 
 enum replaygain {
 	RG_DISABLED,
 	RG_TRACK,
-	RG_ALBUM
-};
-
-struct player_callbacks {
-	int (*get_next)(struct track_info **ti);
+	RG_ALBUM,
+	RG_TRACK_PREFERRED,
+	RG_ALBUM_PREFERRED
 };
 
 struct player_info {
-	pthread_mutex_t mutex;
-
 	/* current track */
 	struct track_info *ti;
 
-	/* stream metadata */
-	char metadata[255 * 16 + 1];
-
 	/* status */
 	enum player_status status;
 	int pos;
-
-	/* volume */
-	int vol_left;
-	int vol_right;
+	int current_bitrate;
 
 	int buffer_fill;
 	int buffer_size;
@@ -79,21 +69,20 @@ struct player_info {
 	unsigned int status_changed : 1;
 	unsigned int position_changed : 1;
 	unsigned int buffer_fill_changed : 1;
-	unsigned int vol_changed : 1;
 };
 
+extern char player_metadata[255 * 16 + 1];
 extern struct player_info player_info;
 extern int player_cont;
+extern int player_repeat_current;
 extern enum replaygain replaygain;
 extern int replaygain_limit;
 extern double replaygain_preamp;
+extern int soft_vol;
+extern int soft_vol_l;
+extern int soft_vol_r;
 
-/* defined in output.c */
-extern int volume_max;
-
-void player_load_plugins(void);
-
-void player_init(const struct player_callbacks *callbacks);
+void player_init(void);
 void player_exit(void);
 
 /* set current file */
@@ -102,30 +91,30 @@ void player_set_file(struct track_info *ti);
 /* set current file and start playing */
 void player_play_file(struct track_info *ti);
 
+/* update track info */
+void player_file_changed(struct track_info *ti);
+
 void player_play(void);
 void player_stop(void);
 void player_pause(void);
-void player_seek(double offset, int relative);
+void player_pause_playback(void);
+void player_seek(double offset, int relative, int start_playing);
 void player_set_op(const char *name);
-char *player_get_op(void);
 void player_set_buffer_chunks(unsigned int nr_chunks);
 int player_get_buffer_chunks(void);
-void player_set_buffer_seconds(unsigned int seconds);
-int player_get_fileinfo(const char *filename, int *duration, struct keyval **comments);
+void player_info_snapshot(void);
 
-int player_get_volume(int *left, int *right);
-int player_set_volume(int left, int right);
+void player_set_soft_volume(int l, int r);
 void player_set_soft_vol(int soft);
 void player_set_rg(enum replaygain rg);
 void player_set_rg_limit(int limit);
 void player_set_rg_preamp(double db);
 
-int player_set_op_option(unsigned int id, const char *val);
-int player_get_op_option(unsigned int id, char **val);
-int player_for_each_op_option(void (*callback)(unsigned int id, const char *key));
-void player_dump_plugins(void);
+#define VF_RELATIVE	0x01
+#define VF_PERCENTAGE	0x02
+int player_set_vol(int l, int lf, int r, int rf);
 
-#define player_info_lock() cmus_mutex_lock(&player_info.mutex)
-#define player_info_unlock() cmus_mutex_unlock(&player_info.mutex)
+void player_metadata_lock(void);
+void player_metadata_unlock(void);
 
 #endif
