@@ -16,6 +16,7 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#include "job.h"
 #include "convert.h"
 #include "ui_curses.h"
 #include "cmdline.h"
@@ -48,6 +49,8 @@
 #include "file.h"
 #include "path.h"
 #include "mixer.h"
+#include "mpris.h"
+#include "locking.h"
 #ifdef HAVE_CONFIG
 #include "config/curses.h"
 #include "config/iconv.h"
@@ -86,30 +89,24 @@ char *tgoto(const char *cap, int col, int row);
 
 /* globals. documented in ui_curses.h */
 
-sig_atomic_t cmus_running = 1;
+volatile sig_atomic_t cmus_running = 1;
 int ui_initialized = 0;
 enum ui_input_mode input_mode = NORMAL_MODE;
 int cur_view = TREE_VIEW;
+int prev_view = -1;
 struct searchable *searchable;
 char *lib_filename = NULL;
 char *lib_ext_filename = NULL;
-char *pl_filename = NULL;
-char *pl_ext_filename = NULL;
 char *play_queue_filename = NULL;
 char *play_queue_ext_filename = NULL;
 char *charset = NULL;
 int using_utf8 = 0;
 
-
 /* ------------------------------------------------------------------------- */
 
 static char *lib_autosave_filename;
-static char *pl_autosave_filename;
 static char *play_queue_autosave_filename;
 
-static int notify_in;
-static int notify_out;
-
 /* shown error message and time stamp
  * error is cleared if it is older than 3s and key was pressed
  */
@@ -121,7 +118,7 @@ static int error_count = 0;
 
 static char *server_address = NULL;
 
-static char print_buffer[512];
+static char print_buffer[1024];
 
 /* destination buffer for utf8_encode_to_buf and utf8_decode */
 static char conv_buffer[512];
@@ -137,17 +134,21 @@ static const char *t_ts;
 static const char *t_fs;
 
 static int tree_win_x = 0;
-static int tree_win_y = 0;
 static int tree_win_w = 0;
 
 static int track_win_x = 0;
-static int track_win_y = 0;
 static int track_win_w = 0;
 
+static int editable_win_x = 0;
+static int editable_win_w = 0;
+static int editable_active = 1;
+
 static int show_cursor;
 static int cursor_x;
 static int cursor_y;
 
+static const int default_esc_delay = 25;
+
 static char *title_buf = NULL;
 
 enum {
@@ -171,6 +172,8 @@ enum {
 	CURSED_ERROR,
 	CURSED_INFO,
 
+	CURSED_TRACKWIN_ALBUM,
+
 	NR_CURSED
 };
 
@@ -193,7 +196,9 @@ static unsigned char cursed_to_bg_idx[NR_CURSED] = {
 	COLOR_TITLELINE_BG,
 	COLOR_WIN_BG,
 	COLOR_CMDLINE_BG,
-	COLOR_CMDLINE_BG
+	COLOR_CMDLINE_BG,
+
+	COLOR_TRACKWIN_ALBUM_BG,
 };
 
 static unsigned char cursed_to_fg_idx[NR_CURSED] = {
@@ -215,17 +220,19 @@ static unsigned char cursed_to_fg_idx[NR_CURSED] = {
 	COLOR_TITLELINE_FG,
 	COLOR_WIN_DIR,
 	COLOR_ERROR,
-	COLOR_INFO
+	COLOR_INFO,
+
+	COLOR_TRACKWIN_ALBUM_FG,
 };
 
 static unsigned char cursed_to_attr_idx[NR_CURSED] = {
 	COLOR_WIN_ATTR,
-	COLOR_WIN_ATTR,
+	COLOR_WIN_CUR_ATTR,
 	COLOR_WIN_INACTIVE_SEL_ATTR,
 	COLOR_WIN_INACTIVE_CUR_SEL_ATTR,
 
 	COLOR_WIN_ATTR,
-	COLOR_WIN_ATTR,
+	COLOR_WIN_CUR_ATTR,
 	COLOR_WIN_SEL_ATTR,
 	COLOR_WIN_CUR_SEL_ATTR,
 
@@ -237,7 +244,9 @@ static unsigned char cursed_to_attr_idx[NR_CURSED] = {
 	COLOR_TITLELINE_ATTR,
 	COLOR_WIN_ATTR,
 	COLOR_CMDLINE_ATTR,
-	COLOR_CMDLINE_ATTR
+	COLOR_CMDLINE_ATTR,
+
+	COLOR_TRACKWIN_ALBUM_ATTR,
 };
 
 /* index is CURSED_*, value is fucking color pair */
@@ -258,6 +267,7 @@ enum {
 	TF_COMMENT,
 	TF_DURATION,
 	TF_DURATION_SEC,
+	TF_ALBUMDURATION,
 	TF_BITRATE,
 	TF_CODEC,
 	TF_CODEC_PROFILE,
@@ -295,6 +305,7 @@ enum {
 	TF_FOLLOW,
 	TF_SHUFFLE,
 	TF_PLAYLISTMODE,
+	TF_BPM,
 
 	NR_TFS
 };
@@ -314,6 +325,7 @@ static struct format_option track_fopts[NR_TFS + 1] = {
 	DEF_FO_STR('c', "comment", 0),
 	DEF_FO_TIME('d', "duration", 0),
 	DEF_FO_INT('\0', "duration_sec", 1),
+	DEF_FO_TIME('\0', "albumduration", 0),
 	DEF_FO_INT('\0', "bitrate", 0),
 	DEF_FO_STR('\0', "codec", 0),
 	DEF_FO_STR('\0', "codec_profile", 0),
@@ -351,11 +363,10 @@ static struct format_option track_fopts[NR_TFS + 1] = {
 	DEF_FO_STR('\0', "follow", 0),
 	DEF_FO_STR('\0', "shuffle", 0),
 	DEF_FO_STR('\0', "playlist_mode", 0),
+	DEF_FO_INT('\0', "bpm", 0),
 	DEF_FO_END
 };
 
-static int last_mevent = 0;
-
 int get_track_win_x(void)
 {
 	return track_win_x;
@@ -397,7 +408,7 @@ static void utf8_encode_to_buf(const char *buffer)
 	return;
 fallback:
 #endif
-	n = min(sizeof(conv_buffer) - 1, strlen(buffer));
+	n = min_i(sizeof(conv_buffer) - 1, strlen(buffer));
 	memmove(conv_buffer, buffer, n);
 	conv_buffer[n] = '\0';
 }
@@ -588,7 +599,7 @@ static void fill_track_fopts_track_info(struct track_info *info)
 	fopt_set_double(&track_fopts[TF_RG_TRACK_PEAK], info->rg_track_peak, isnan(info->rg_track_peak));
 	fopt_set_double(&track_fopts[TF_RG_ALBUM_GAIN], info->rg_album_gain, isnan(info->rg_album_gain));
 	fopt_set_double(&track_fopts[TF_RG_ALBUM_PEAK], info->rg_album_peak, isnan(info->rg_album_peak));
-	fopt_set_int(&track_fopts[TF_ORIGINALYEAR], info->date / 10000, info->originaldate <= 0);
+	fopt_set_int(&track_fopts[TF_ORIGINALYEAR], info->originaldate / 10000, info->originaldate <= 0);
 	fopt_set_int(&track_fopts[TF_BITRATE], (int) (info->bitrate / 1000. + 0.5), info->bitrate == -1);
 	fopt_set_str(&track_fopts[TF_CODEC], info->codec);
 	fopt_set_str(&track_fopts[TF_CODEC_PROFILE], info->codec_profile);
@@ -613,6 +624,20 @@ static void fill_track_fopts_track_info(struct track_info *info)
 	} else {
 		fopt_set_str(&track_fopts[TF_FILE], path_basename(filename));
 	}
+	fopt_set_int(&track_fopts[TF_BPM], info->bpm, info->bpm == -1);
+}
+
+static int get_album_length(struct album *album)
+{
+	struct tree_track *track;
+	struct rb_node *tmp;
+	int duration = 0;
+
+	rb_for_each_entry(track, tmp, &album->track_root, tree_node) {
+		duration += tree_track_info(track)->duration;
+	}
+
+	return duration;
 }
 
 static void fill_track_fopts_album(struct album *album)
@@ -622,12 +647,14 @@ static void fill_track_fopts_album(struct album *album)
 	fopt_set_str(&track_fopts[TF_ALBUMARTIST], album->artist->name);
 	fopt_set_str(&track_fopts[TF_ARTIST], album->artist->name);
 	fopt_set_str(&track_fopts[TF_ALBUM], album->name);
+	fopt_set_time(&track_fopts[TF_ALBUMDURATION], get_album_length(album), 0);
 }
 
 static void fill_track_fopts_artist(struct artist *artist)
 {
 	const char *name = display_artist_sort_name ? artist_sort_name(artist) : artist->name;
 	fopt_set_str(&track_fopts[TF_ARTIST], name);
+	fopt_set_str(&track_fopts[TF_ALBUMARTIST], name);
 }
 
 const struct format_option *get_global_fopts(void)
@@ -643,10 +670,8 @@ const struct format_option *get_global_fopts(void)
 	int buffer_fill, vol, vol_left, vol_right;
 	int duration = -1;
 
-	editable_lock();
 	fopt_set_time(&track_fopts[TF_TOTAL], play_library ? lib_editable.total_time :
-			pl_editable.total_time, 0);
-	editable_unlock();
+			pl_playing_total_time(), 0);
 
 	fopt_set_str(&track_fopts[TF_FOLLOW], follow_strs[follow]);
 	fopt_set_str(&track_fopts[TF_REPEAT], repeat_strs[repeat]);
@@ -726,7 +751,7 @@ static void print_tree(struct window *win, int row, struct iter *iter)
 	pos = strlen(print_buffer);
 	print_buffer[pos++] = ' ';
 	print_buffer[pos++] = 0;
-	dump_print_buffer(tree_win_y + row + 1, tree_win_x);
+	dump_print_buffer(row + 1, tree_win_x);
 }
 
 static void print_track(struct window *win, int row, struct iter *iter)
@@ -745,18 +770,16 @@ static void print_track(struct window *win, int row, struct iter *iter)
 		int pos;
 		struct fp_len len;
 
-		/* FIXME:
-		 * replace A_BOLD by something useful */
-		bkgdset(A_BOLD);
+		bkgdset(pairs[CURSED_TRACKWIN_ALBUM]);
 
 		fill_track_fopts_album(album);
 
 		len = format_print(print_buffer, track_win_w, track_win_album_format, track_fopts);
-		dump_print_buffer(track_win_y + row + 1, track_win_x);
+		dump_print_buffer(row + 1, track_win_x);
 
 		bkgdset(pairs[CURSED_SEPARATOR]);
 		for(pos = track_win_x + len.llen; pos < COLS - len.rlen; ++pos)
-			(void) mvaddch(track_win_y + row + 1, pos, ACS_HLINE);
+			(void) mvaddch(row + 1, pos, ACS_HLINE);
 
 		return;
 	}
@@ -783,7 +806,7 @@ static void print_track(struct window *win, int row, struct iter *iter)
 		format = track_win_alt_format;
 	}
 	format_print(print_buffer, track_win_w, format, track_fopts);
-	dump_print_buffer(track_win_y + row + 1, track_win_x);
+	dump_print_buffer(row + 1, track_win_x);
 }
 
 /* used by print_editable only */
@@ -802,11 +825,11 @@ static void print_editable(struct window *win, int row, struct iter *iter)
 	selected = iters_equal(iter, &sel);
 
 	if (selected) {
-		cursor_x = 0;
+		cursor_x = editable_win_x;
 		cursor_y = 1 + row;
 	}
 
-	active = 1;
+	active = editable_active;
 	if (!selected && track->marked) {
 		selected = 1;
 		active = 0;
@@ -823,8 +846,8 @@ static void print_editable(struct window *win, int row, struct iter *iter)
 	} else if (*list_win_alt_format) {
 		format = list_win_alt_format;
 	}
-	format_print(print_buffer, COLS, format, track_fopts);
-	dump_print_buffer(row + 1, 0);
+	format_print(print_buffer, editable_win_w, format, track_fopts);
+	dump_print_buffer(row + 1, editable_win_x);
 }
 
 static void print_browser(struct window *win, int row, struct iter *iter)
@@ -932,9 +955,9 @@ static void print_help(struct window *win, int row, struct iter *iter)
 		break;
 	case HE_BOUND:
 		snprintf(buf, sizeof(buf), " %-8s %-14s %s",
-			key_context_names[e->binding->ctx],
-			e->binding->key->name,
-			e->binding->cmd);
+				key_context_names[e->binding->ctx],
+				e->binding->key->name,
+				e->binding->cmd);
 		break;
 	case HE_UNBOUND:
 		snprintf(buf, sizeof(buf), " %s", e->command->name);
@@ -942,7 +965,8 @@ static void print_help(struct window *win, int row, struct iter *iter)
 	case HE_OPTION:
 		opt = e->option;
 		snprintf(buf, sizeof(buf), " %-29s ", opt->name);
-		opt->get(opt->id, buf + strlen(buf));
+		size_t len = strlen(buf);
+		opt->get(opt->data, buf + len, sizeof(buf) - len);
 		break;
 	}
 	pos = format_str(print_buffer, buf, COLS - 1);
@@ -989,8 +1013,8 @@ static void update_window(struct window *win, int x, int y, int w, const char *t
 
 static void update_tree_window(void)
 {
-	update_window(lib_tree_win, tree_win_x, tree_win_y,
-			tree_win_w, "Artist / Album", print_tree);
+	update_window(lib_tree_win, tree_win_x, 0, tree_win_w,
+			"Artist / Album", print_tree);
 }
 
 static void update_track_window(void)
@@ -1000,8 +1024,54 @@ static void update_track_window(void)
 	/* it doesn't matter what format options we use because the format
 	 * string does not contain any format charaters */
 	format_print(title, track_win_w - 2, "Track%=Library", track_fopts);
-	update_window(lib_track_win, track_win_x, track_win_y,
-			track_win_w, title, print_track);
+	update_window(lib_track_win, track_win_x, 0, track_win_w, title,
+			print_track);
+}
+
+static void print_pl_list(struct window *win, int row, struct iter *iter)
+{
+	struct pl_list_info info;
+
+	pl_list_iter_to_info(iter, &info);
+
+	bkgdset(pairs[(info.active<<2) | (info.selected<<1) | info.current]);
+
+	const char *prefix = "   ";
+	if (info.marked)
+		prefix = " * ";
+	size_t prefix_w = strlen(prefix);
+	format_str(print_buffer, prefix, prefix_w);
+
+	if (tree_win_w >= prefix_w)
+		format_str(print_buffer + prefix_w, info.name,
+				tree_win_w - prefix_w);
+
+	dump_print_buffer(row + 1, 0);
+}
+
+static void update_pl_list(struct window *win)
+{
+	update_window(win, tree_win_x, 0, tree_win_w, "Playlist",
+			print_pl_list);
+}
+
+static void update_pl_tracks(struct window *win)
+{
+	char title[512];
+
+	editable_win_x = track_win_x;
+	editable_win_w = track_win_w;
+	editable_active = pl_get_cursor_in_track_window();
+
+	get_global_fopts();
+	fopt_set_time(&track_fopts[TF_TOTAL], pl_visible_total_time(), 0);
+
+	format_print(title, track_win_w - 2, "Track%=%{total}", track_fopts);
+	update_window(win, track_win_x, 0, track_win_w, title, print_editable);
+
+	editable_active = 1;
+	editable_win_x = 0;
+	editable_win_w = COLS;
 }
 
 static const char *pretty(const char *path)
@@ -1046,9 +1116,10 @@ static void update_editable_window(struct editable *e, const char *title, const
 	}
 	pos = strlen(buf);
 	snprintf(buf + pos, sizeof(buf) - pos, " %s%s",
-			sorted_names[e->sort_str[0] != 0], e->sort_str);
+			sorted_names[e->shared->sort_str[0] != 0],
+			e->shared->sort_str);
 
-	update_window(e->win, 0, 0, COLS, buf, &print_editable);
+	update_window(e->shared->win, 0, 0, COLS, buf, &print_editable);
 }
 
 static void update_sorted_window(void)
@@ -1057,12 +1128,6 @@ static void update_sorted_window(void)
 	update_editable_window(&lib_editable, "Library", lib_filename);
 }
 
-static void update_pl_window(void)
-{
-	current_track = pl_cur_track;
-	update_editable_window(&pl_editable, "Playlist", pl_filename);
-}
-
 static void update_play_queue_window(void)
 {
 	current_track = NULL;
@@ -1106,6 +1171,13 @@ static void draw_separator(void)
 		(void) mvaddch(row, tree_win_w, ACS_VLINE);
 }
 
+static void update_pl_view(int full)
+{
+	current_track = pl_get_playing_track();
+	pl_draw(update_pl_list, update_pl_tracks, full);
+	draw_separator();
+}
+
 static void do_update_view(int full)
 {
 	cursor_x = -1;
@@ -1113,30 +1185,22 @@ static void do_update_view(int full)
 
 	switch (cur_view) {
 	case TREE_VIEW:
-		editable_lock();
 		if (full || lib_tree_win->changed)
 			update_tree_window();
 		if (full || lib_track_win->changed)
 			update_track_window();
-		editable_unlock();
 		draw_separator();
 		update_filterline();
 		break;
 	case SORTED_VIEW:
-		editable_lock();
 		update_sorted_window();
-		editable_unlock();
 		update_filterline();
 		break;
 	case PLAYLIST_VIEW:
-		editable_lock();
-		update_pl_window();
-		editable_unlock();
+		update_pl_view(full);
 		break;
 	case QUEUE_VIEW:
-		editable_lock();
 		update_play_queue_window();
-		editable_unlock();
 		break;
 	case BROWSER_VIEW:
 		update_browser_window();
@@ -1152,21 +1216,12 @@ static void do_update_view(int full)
 
 static void do_update_statusline(void)
 {
-	player_info_lock();
 	format_print(print_buffer, COLS, statusline_format, get_global_fopts());
-
-	char *msg = player_info.error_msg;
-	player_info.error_msg = NULL;
-
-	player_info_unlock();
-
 	bkgdset(pairs[CURSED_STATUSLINE]);
 	dump_print_buffer(LINES - 2, 0);
 
-	if (msg) {
-		error_msg("%s", msg);
-		free(msg);
-	}
+	if (player_info.error_msg)
+		error_msg("%s", player_info.error_msg);
 }
 
 static void dump_buffer(const char *buffer)
@@ -1292,7 +1347,6 @@ static void set_title(const char *title)
 static void do_update_titleline(void)
 {
 	bkgdset(pairs[CURSED_TITLELINE]);
-	player_info_lock();
 	if (player_info.ti) {
 		int i, use_alt_format = 0;
 		char *wtitle;
@@ -1351,7 +1405,6 @@ static void do_update_titleline(void)
 
 		set_title("cmus " VERSION);
 	}
-	player_info_unlock();
 }
 
 static int cmdline_cursor_column(void)
@@ -1419,13 +1472,12 @@ static void post_update(void)
 	}
 }
 
-/* lock player_info! */
-const char *get_stream_title(void)
+static const char *get_stream_title_locked(void)
 {
 	static char stream_title[255 * 16 + 1];
 	char *ptr, *title;
 
-	ptr = strstr(player_info.metadata, "StreamTitle='");
+	ptr = strstr(player_metadata, "StreamTitle='");
 	if (ptr == NULL)
 		return NULL;
 	ptr += 13;
@@ -1441,6 +1493,14 @@ const char *get_stream_title(void)
 	return NULL;
 }
 
+const char *get_stream_title(void)
+{
+	player_metadata_lock();
+	const char *rv = get_stream_title_locked();
+	player_metadata_unlock();
+	return rv;
+}
+
 void update_titleline(void)
 {
 	curs_set(0);
@@ -1635,19 +1695,20 @@ void set_view(int view)
 	if (view == cur_view)
 		return;
 
+	prev_view = cur_view;
 	cur_view = view;
 	switch (cur_view) {
 	case TREE_VIEW:
 		searchable = tree_searchable;
 		break;
 	case SORTED_VIEW:
-		searchable = lib_editable.searchable;
+		searchable = lib_editable.shared->searchable;
 		break;
 	case PLAYLIST_VIEW:
-		searchable = pl_editable.searchable;
+		searchable = pl_get_searchable();
 		break;
 	case QUEUE_VIEW:
-		searchable = pq_editable.searchable;
+		searchable = pq_editable.shared->searchable;
 		break;
 	case BROWSER_VIEW:
 		searchable = browser_searchable;
@@ -1743,7 +1804,6 @@ static void spawn_status_program(void)
 	if (status_display_program == NULL || status_display_program[0] == 0)
 		return;
 
-	player_info_lock();
 	status = player_info.status;
 	if (status == PLAYER_STATUS_PLAYING && player_info.ti && is_http_url(player_info.ti->filename))
 		stream_title = get_stream_title();
@@ -1755,8 +1815,8 @@ static void spawn_status_program(void)
 	argv[i++] = xstrdup(player_status_names[status]);
 	if (player_info.ti) {
 		static const char *keys[] = {
-			"artist", "album", "discnumber", "tracknumber", "title", "date",
-			"musicbrainz_trackid", NULL
+			"artist", "albumartist", "album", "discnumber", "tracknumber", "title",
+			"date",	"musicbrainz_trackid", NULL
 		};
 		int j;
 
@@ -1797,7 +1857,6 @@ static void spawn_status_program(void)
 		}
 	}
 	argv[i++] = NULL;
-	player_info_unlock();
 
 	if (spawn(argv, NULL, 0) == -1)
 		error_msg("couldn't run `%s': %s", status_display_program, strerror(errno));
@@ -1805,7 +1864,7 @@ static void spawn_status_program(void)
 		free(argv[i]);
 }
 
-static sig_atomic_t ctrl_c_pressed = 0;
+static volatile sig_atomic_t ctrl_c_pressed = 0;
 
 static void sig_int(int sig)
 {
@@ -1817,7 +1876,7 @@ static void sig_shutdown(int sig)
 	cmus_running = 0;
 }
 
-static sig_atomic_t needs_to_resize = 1;
+static volatile sig_atomic_t needs_to_resize = 1;
 
 static void sig_winch(int sig)
 {
@@ -1844,9 +1903,7 @@ static void resize_tree_view(int w, int h)
 	if (track_win_w < 8)
 		track_win_w = 8;
 	tree_win_x = 0;
-	tree_win_y = 0;
 	track_win_x = tree_win_w + 1;
-	track_win_y = 0;
 
 	h--;
 	window_set_nr_rows(lib_tree_win, h);
@@ -1870,21 +1927,20 @@ static void update(void)
 #if HAVE_RESIZETERM
 			resizeterm(lines, columns);
 #endif
+			editable_win_w = COLS;
 			w = COLS;
 			h = LINES - 3;
 			if (w < 16)
 				w = 16;
-			if (h < 8)
-				h = 8;
-			editable_lock();
+			if (h < 2)
+				h = 2;
 			resize_tree_view(w, h);
-			window_set_nr_rows(lib_editable.win, h - 1);
-			window_set_nr_rows(pl_editable.win, h - 1);
-			window_set_nr_rows(pq_editable.win, h - 1);
+			window_set_nr_rows(lib_editable.shared->win, h - 1);
+			pl_set_nr_rows(h - 1);
+			window_set_nr_rows(pq_editable.shared->win, h - 1);
 			window_set_nr_rows(filters_win, h - 1);
 			window_set_nr_rows(help_win, h - 1);
 			window_set_nr_rows(browser_win, h - 1);
-			editable_unlock();
 			needs_title_update = 1;
 			needs_status_update = 1;
 			needs_command_update = 1;
@@ -1893,39 +1949,35 @@ static void update(void)
 		refresh();
 	}
 
-	player_info_lock();
-	editable_lock();
+	if (player_info.status_changed)
+		mpris_playback_status_changed();
+
+	if (player_info.file_changed || player_info.metadata_changed)
+		mpris_metadata_changed();
 
 	needs_spawn = player_info.status_changed || player_info.file_changed ||
 		player_info.metadata_changed;
 
 	if (player_info.file_changed) {
-		player_info.file_changed = 0;
 		needs_title_update = 1;
 		needs_status_update = 1;
 	}
-	if (player_info.metadata_changed) {
-		player_info.metadata_changed = 0;
+	if (player_info.metadata_changed)
 		needs_title_update = 1;
-	}
-	if (player_info.position_changed || player_info.status_changed) {
-		player_info.position_changed = 0;
-		player_info.status_changed = 0;
-
+	if (player_info.position_changed || player_info.status_changed)
 		needs_status_update = 1;
-	}
 	switch (cur_view) {
 	case TREE_VIEW:
 		needs_view_update += lib_tree_win->changed || lib_track_win->changed;
 		break;
 	case SORTED_VIEW:
-		needs_view_update += lib_editable.win->changed;
+		needs_view_update += lib_editable.shared->win->changed;
 		break;
 	case PLAYLIST_VIEW:
-		needs_view_update += pl_editable.win->changed;
+		needs_view_update += pl_needs_redraw();
 		break;
 	case QUEUE_VIEW:
-		needs_view_update += pq_editable.win->changed;
+		needs_view_update += pq_editable.shared->win->changed;
 		break;
 	case BROWSER_VIEW:
 		needs_view_update += browser_win->changed;
@@ -1940,16 +1992,12 @@ static void update(void)
 
 	/* total time changed? */
 	if (play_library) {
-		needs_status_update += lib_editable.win->changed;
-		lib_editable.win->changed = 0;
+		needs_status_update += lib_editable.shared->win->changed;
+		lib_editable.shared->win->changed = 0;
 	} else {
-		needs_status_update += pl_editable.win->changed;
-		pl_editable.win->changed = 0;
+		needs_status_update += pl_needs_redraw();
 	}
 
-	editable_unlock();
-	player_info_unlock();
-
 	if (needs_spawn)
 		spawn_status_program();
 
@@ -2013,10 +2061,13 @@ static void handle_key(int key)
 static void handle_mouse(MEVENT *event)
 {
 #if NCURSES_MOUSE_VERSION <= 1
+	static int last_mevent;
+
 	if ((last_mevent & BUTTON1_PRESSED) && (event->bstate & REPORT_MOUSE_POSITION))
 		event->bstate = BUTTON1_RELEASED;
 	last_mevent = event->bstate;
 #endif
+
 	clear_error();
 	if (input_mode == NORMAL_MODE) {
 		normal_mode_mouse(event);
@@ -2054,7 +2105,9 @@ static void u_getch(void)
 
 	/* escape sequence */
 	if (key == 0x1B) {
+		cbreak();
 		int e_key = getch();
+		halfdelay(5);
 		if (e_key != ERR && e_key != 0) {
 			handle_escape(e_key);
 			return;
@@ -2088,18 +2141,16 @@ static void u_getch(void)
 	handle_ch(u);
 }
 
-void ui_curses_notify(void)
-{
-	char c;
-	if (write(notify_in, &c, 1) == -1) {
-		d_print("write failed: %s\n", strerror(errno));
-	}
-}
-
 static void main_loop(void)
 {
 	int rc, fd_high;
 
+#define SELECT_ADD_FD(fd) do {\
+	FD_SET((fd), &set); \
+	if ((fd) > fd_high) \
+		fd_high = (fd); \
+} while(0)
+
 	fd_high = server_socket;
 	while (cmus_running) {
 		fd_set set;
@@ -2110,10 +2161,12 @@ static void main_loop(void)
 		struct list_head *item;
 		struct client *client;
 
+		player_info_snapshot();
+
 		update();
 
 		/* Timeout must be so small that screen updates seem instant.
-		 * Only affects changes done in other threads (worker, player).
+		 * Only affects changes done in other threads (player).
 		 *
 		 * Too small timeout makes window updates too fast (wastes CPU).
 		 *
@@ -2123,28 +2176,20 @@ static void main_loop(void)
 		tv.tv_sec = 0;
 		tv.tv_usec = 0;
 
-		player_info_lock();
 		if (player_info.status == PLAYER_STATUS_PLAYING) {
 			// player position updates need to be fast
 			tv.tv_usec = 100e3;
 		}
-		player_info_unlock();
-
-		if (!tv.tv_usec && worker_has_job(JOB_TYPE_ANY)) {
-			// playlist is loading. screen needs to be updated
-			tv.tv_usec = 250e3;
-		}
 
 		FD_ZERO(&set);
-		FD_SET(0, &set);
-		FD_SET(notify_out, &set);
-		if (notify_out > fd_high)
-			fd_high = notify_out;
-		FD_SET(server_socket, &set);
+		SELECT_ADD_FD(0);
+		SELECT_ADD_FD(job_fd);
+		SELECT_ADD_FD(cmus_next_track_request_fd);
+		SELECT_ADD_FD(server_socket);
+		if (mpris_fd != -1)
+			SELECT_ADD_FD(mpris_fd);
 		list_for_each_entry(client, &client_head, node) {
-			FD_SET(client->fd, &set);
-			if (client->fd > fd_high)
-				fd_high = client->fd;
+			SELECT_ADD_FD(client->fd);
 		}
 		if (!soft_vol) {
 			nr_fds = mixer_get_fds(fds);
@@ -2155,24 +2200,20 @@ static void main_loop(void)
 			}
 			for (i = 0; i < nr_fds; i++) {
 				BUG_ON(fds[i] <= 0);
-				FD_SET(fds[i], &set);
-				if (fds[i] > fd_high)
-					fd_high = fds[i];
+				SELECT_ADD_FD(fds[i]);
 			}
 		}
 
-		if (tv.tv_usec) {
-			rc = select(fd_high + 1, &set, NULL, NULL, &tv);
-		} else {
-			rc = select(fd_high + 1, &set, NULL, NULL, NULL);
-		}
+		rc = select(fd_high + 1, &set, NULL, NULL, tv.tv_usec ? &tv : NULL);
 		if (poll_mixer) {
 			int ol = volume_l;
 			int or = volume_r;
 
 			mixer_read_volume();
-			if (ol != volume_l || or != volume_r)
+			if (ol != volume_l || or != volume_r) {
+				mpris_volume_changed();
 				update_statusline();
+			}
 
 		}
 		if (rc <= 0) {
@@ -2188,6 +2229,7 @@ static void main_loop(void)
 			if (FD_ISSET(fds[i], &set)) {
 				d_print("vol changed\n");
 				mixer_read_volume();
+				mpris_volume_changed();
 				update_statusline();
 			}
 		}
@@ -2207,41 +2249,17 @@ static void main_loop(void)
 		if (FD_ISSET(0, &set))
 			u_getch();
 
-		if (FD_ISSET(notify_out, &set)) {
-			char buf[128];
-			if (read(notify_out, buf, sizeof(buf)) == -1) {
-				d_print("read failed: %s\n", strerror(errno));
-			}
-		}
+		if (mpris_fd != -1 && FD_ISSET(mpris_fd, &set))
+			mpris_process();
+
+		if (FD_ISSET(job_fd, &set))
+			job_handle();
+
+		if (FD_ISSET(cmus_next_track_request_fd, &set))
+			cmus_provide_next_track();
 	}
 }
 
-static int get_next(struct track_info **ti)
-{
-	struct track_info *info;
-
-	editable_lock();
-	info = play_queue_remove();
-	if (info == NULL) {
-		if (play_library) {
-			info = lib_goto_next();
-		} else {
-			info = pl_goto_next();
-		}
-	}
-	editable_unlock();
-
-	if (info == NULL)
-		return -1;
-
-	*ti = info;
-	return 0;
-}
-
-static const struct player_callbacks player_callbacks = {
-	.get_next = get_next
-};
-
 static void init_curses(void)
 {
 	struct sigaction act;
@@ -2269,18 +2287,11 @@ static void init_curses(void)
 	sigaction(SIGWINCH, &act, NULL);
 
 	initscr();
-
-	/* turn off kb buffering */
-	cbreak();
-
+	nodelay(stdscr, TRUE);
 	keypad(stdscr, TRUE);
-
-	/* wait max 5 * 0.1 s if there are no keys available
-	 * doesn't really matter because we use select()
-	 */
 	halfdelay(5);
-
 	noecho();
+
 	if (has_colors()) {
 #if HAVE_USE_DEFAULT_COLORS
 		start_color();
@@ -2328,32 +2339,29 @@ static void init_curses(void)
 		}
 	}
 	update_mouse();
-}
 
-static void notify_init(void)
-{
-	int fildes[2];
-	BUG_ON(pipe(fildes) == -1);
-	notify_out = fildes[0];
-	notify_in = fildes[1];
-	int flags = fcntl(notify_in, F_GETFL, 0);
-	fcntl(notify_in, F_SETFL, flags | O_NONBLOCK);
+	if (!getenv("ESCDELAY")) {
+		set_escdelay(default_esc_delay);
+	}
 }
 
 static void init_all(void)
 {
+	main_thread = pthread_self();
+	cmus_track_request_init();
+
 	server_init(server_address);
 
 	/* does not select output plugin */
-	player_init(&player_callbacks);
+	player_init();
 
 	/* plugins have been loaded so we know what plugin options are available */
 	options_add();
 
 	lib_init();
 	searchable = tree_searchable;
-	pl_init();
 	cmus_init();
+	pl_init();
 	browser_init();
 	filters_init();
 	help_init();
@@ -2363,6 +2371,8 @@ static void init_all(void)
 
 	/* almost everything must be initialized now */
 	options_load();
+	if (mpris)
+		mpris_init();
 
 	/* finally we can set the output plugin */
 	player_set_op(output_plugin);
@@ -2370,9 +2380,7 @@ static void init_all(void)
 		mixer_open();
 
 	lib_autosave_filename = xstrjoin(cmus_config_dir, "/lib.pl");
-	pl_autosave_filename = xstrjoin(cmus_config_dir, "/playlist.pl");
 	play_queue_autosave_filename = xstrjoin(cmus_config_dir, "/queue.pl");
-	pl_filename = xstrdup(pl_autosave_filename);
 	lib_filename = xstrdup(lib_autosave_filename);
 
 	if (error_count) {
@@ -2390,11 +2398,16 @@ static void init_all(void)
 
 	if (resume_cmus) {
 		resume_load();
-		cmus_add(play_queue_append, play_queue_autosave_filename, FILE_TYPE_PL, JOB_TYPE_QUEUE, 0);
+		cmus_add(play_queue_append, play_queue_autosave_filename,
+				FILE_TYPE_PL, JOB_TYPE_QUEUE, 0, NULL);
+	} else {
+		set_view(start_view);
 	}
 
-	cmus_add(pl_add_track, pl_autosave_filename, FILE_TYPE_PL, JOB_TYPE_PL, 0);
-	cmus_add(lib_add_track, lib_autosave_filename, FILE_TYPE_PL, JOB_TYPE_LIB, 0);
+	cmus_add(lib_add_track, lib_autosave_filename, FILE_TYPE_PL,
+			JOB_TYPE_LIB, 0, NULL);
+
+	worker_start();
 }
 
 static void exit_all(void)
@@ -2408,10 +2421,11 @@ static void exit_all(void)
 	server_exit();
 	cmus_exit();
 	if (resume_cmus)
-		cmus_save(play_queue_for_each, play_queue_autosave_filename);
-	cmus_save(lib_for_each, lib_autosave_filename);
-	cmus_save(pl_for_each, pl_autosave_filename);
+		cmus_save(play_queue_for_each, play_queue_autosave_filename,
+				NULL);
+	cmus_save(lib_for_each, lib_autosave_filename, NULL);
 
+	pl_exit();
 	player_exit();
 	op_exit_plugins();
 	commands_exit();
@@ -2419,6 +2433,7 @@ static void exit_all(void)
 	filters_exit();
 	help_exit();
 	browser_exit();
+	mpris_free();
 }
 
 enum {
@@ -2475,7 +2490,7 @@ int main(int argc, char *argv[])
 		case FLAG_VERSION:
 			printf("cmus " VERSION
 			       "\nCopyright 2004-2006 Timo Hirvonen"
-			       "\nCopyright 2008-2013 Various Authors\n");
+			       "\nCopyright 2008-2016 Various Authors\n");
 			return 0;
 		case FLAG_PLUGINS:
 			list_plugins = 1;
@@ -2515,7 +2530,6 @@ int main(int argc, char *argv[])
 		op_dump_plugins();
 		return 0;
 	}
-	notify_init();
 	init_all();
 	main_loop();
 	exit_all();
