@@ -1,6 +1,7 @@
-/* 
+/*
+ * Copyright 2008-2013 Various Authors
  * Copyright 2004-2005 Timo Hirvonen
- * 
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
@@ -10,11 +11,9 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
- * 02111-1307, USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "xmalloc.h"
@@ -32,18 +31,17 @@ void malloc_fail(void)
 	exit(42);
 }
 
+#ifndef HAVE_STRNDUP
 char *xstrndup(const char *str, size_t n)
 {
-	int len;
+	size_t len;
 	char *s;
 
-	len = strlen(str);
-	if (n > len)
-		n = len;
-	s = malloc(n + 1);
-	if (unlikely(s == NULL))
-		malloc_fail();
-	memcpy(s, str, n);
-	s[n] = 0;
+	for (len = 0; len < n && str[len]; len++)
+		;
+	s = xmalloc(len + 1);
+	memcpy(s, str, len);
+	s[len] = 0;
 	return s;
 }
+#endif
